import sys
import importlib
import os
from pathlib import Path
from lolpop.utils import common_utils as utils
from metaflow import cli, Flow

def get_flow_class(flow_file_path, flow_class):
    #add the parent directory of this file into the classpath
    parent_dir = os.path.dirname(flow_file_path)
    module_name = Path(flow_file_path).stem
    sys.path.append(str(parent_dir))

    #import this file and load the metaflow class in this file
    mod = importlib.import_module(module_name)
    mod_cl = getattr(mod, flow_class)

    return mod_cl 

def load_flow(mod_cl, self_obj, plugin_paths_file, **kwargs):

    #get all child plugin paths and write to file
    #we have to do this because we save the lolpop component as
    #part of metaflow and when it loads up it will fail if plugins
    #are used and not on the system path. This ensures that they are.
    plugin_paths = utils.get_all_plugin_paths(self_obj)
    with open(plugin_paths_file, "w") as f:
        f.write("|".join([str(x) for x in plugin_paths]))

    #load up the metaflow flow.
    flow = mod_cl(use_cli=False, lolpop=self_obj, **kwargs)

    return flow

def run_flow(flow, command, flow_file_path, plugin_paths_file):
    try:
        #now call the main metaflow workflow
        cli.main(flow, args=[command], entrypoint=[sys.executable, flow_file_path])
    except SystemExit:  # metaflow always ends by calling systemexit. To avoid exiting the program, catch it and pass
        pass 
    finally:  
        # remove dir from system path to be safe
        sys.path.remove(str(os.path.dirname(flow_file_path)))
        os.remove(plugin_paths_file)


def get_latest_run(flow_class):
    flow = Flow(flow_class)
    run = flow.latest_run

    return run 

def get_run_artifacts(run, keys, flow_class): 
    if run.successful:
        artifacts = run.data._artifacts
        artifacts_out = [artifacts.get(x) for x in keys]
        return [x.data for x in artifacts_out if x is not None]
    else:
        raise Exception("Metaflow run %s for flow %s failed. Unable to retrieve artifacts." % (run.id, flow_class))


def load_plugins(plugin_paths_file):
    #read the plugin paths from file
    plugin_paths = []
    with open(plugin_paths_file) as f:
        plugin_paths = [Path(x) for x in f.read().split("|") if len(x)>0]

    #load plugins. This ensures that self.lolpop is always populated
    if len(plugin_paths) > 0: 
        _ = utils.load_plugins(plugin_paths)



from pkgutil import extend_path

#allows extensions to work when doing local dev
__path__ = extend_path(__path__, __name__)

def __map_extensions__():
    import os
    from importlib import import_module
    import warnings

    warnings.filterwarnings("ignore")
    #get current directory and all subdirectors. These represent the resource types
    for path in __path__:
        subdirs = ["%s/%s" % (path, x) for x in os.listdir(path)
                if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
        for subdir in subdirs:
            files = [x for x in os.listdir(subdir)
                    if ((x[0] != "_") and (x[0] != ".") and (".py" in x))]

            #from each file, import all classes and register them in the global namespace.
            for file in files:
                try:
                    subdir_arr = subdir.split("/")
                    file_name = file[:-3]
                    module = import_module(
                        "lolpop.cli.extensions.%s.%s" % (subdir_arr[-1],file_name))
                    globals().update({file_name: module})
                except Exception as e:
                    pass

    warnings.resetwarnings()
__map_extensions__()



from lolpop.component.base_component import BaseComponent
from typing import Any 

class BaseModelExplainer(BaseComponent): 
    __REQUIRED_CONF__ = {
        "config" : []
    }
    
    def get_feature_importance(self, data, model, *args, **kwargs) -> tuple[Any, Any]: 
        pass 

    def get_explanations(self, data, model, *args, **kwargs) -> Any:
        pass 


from lolpop.component.test_recorder.base_test_recorder import BaseTestRecorder
import itertools 

class LocalTestRecorder(BaseTestRecorder):
    __REQUIRED_CONF__ = {
        "config": []
    }

    test_results = []

    def record_test(self, obj, method, test, test_method, result, msg=None, *args, **kwargs):
        """Records the results of a test.

        Args:
            obj (object): the lolpop integration tested
            method (object): The method in the integration tested
            test (object): The test module used
            test_method (object): The method in the test module used 
            result (bool): The result of the test. True = passed. 
            msg (str, optional): Additiona information returned from the test. Defaults to None.
        """
        self.test_results.append(
            {"method" : "%s.%s" %(obj.name,method.__name__), 
             "test": "%s.%s" %(test, test_method.__name__), 
             "passed": result, "output": msg}
        )

    def print_report(self): 
        """Prints a testing repot. 
        """
        for key, group in itertools.groupby(self.test_results, key=lambda x: x['method']):
            print("Method: %s" %key)
            for g in group: 
                print("\t Test: %s" %g["test"])
                print("\t Passed: %s" %g["passed"])
                if g["output"] is not None: 
                    print("\t Output: %s" %g["output"] )
            





from lolpop.component.base_component import BaseComponent
from typing import Any 

class BaseMetricsTracker(BaseComponent): 

    url = "https://replace.me"

    def log_metric(self, resource, id, value, *args, **kwargs): 
        pass 

    def get_metric(self, resource, id, *args, **kwargs) -> Any:
        pass

    def log_metrics(self, resource, metrics, *args, **kwargs):
        pass

    def copy_metrics(self, from_resource, to_resource, *args, **kwargs):
        pass

    def log_prediction_metrics(self, prediction_resource, predictions, *args, **kwargs):
        pass



from pkgutil import extend_path

#allows extensions to work when doing local dev
__path__ = extend_path(__path__, __name__)

def __map_extensions__():
    from pathlib import Path
    import os
    from importlib import import_module
    from inspect import isclass
    from lolpop.component.base_component import BaseComponent
    from lolpop.pipeline.base_pipeline import BasePipeline
    from lolpop.runner.base_runner import BaseRunner
    import warnings

    warnings.filterwarnings("ignore")
    #get current directory and all subdirectors. These represent the extensions 
    # resource types
    #path = Path(__file__).parent.resolve() #lolpop/extension
    for path in __path__: 
        extensions = ["%s/%s" % (path, x) for x in os.listdir(path)
                    if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
        #iterate through extensions and add each top level 
        for extension in extensions: #lolpop/extension/my_extension
            resources = ["%s/%s" % (extension, x) for x in os.listdir(extension)
                        if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
            for resource in resources: #lolpop/extension/my_extension/{component,pipeline,runner}
                subdirs = ["%s/%s" % (resource, x) for x in os.listdir(resource)
                        if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
                #for each resource type (i.e. subdir), get all resources implemented in that type (i.e. python files in the subdir)
                for subdir in subdirs:
                    files = [x for x in os.listdir(subdir) if (
                        (".py" in x) and ("__" not in x) and (x[0] != "."))]
                    #from each file, import all classes and register them in the global namespace.
                    for file in files:
                        try: 
                            subdir_arr = subdir.split("/")
                            module = import_module("lolpop.extension.%s.%s.%s.%s" % (
                                subdir_arr[-3], subdir_arr[-2], subdir_arr[-1], file[:-3]))
                            classes = [x for x in dir(module) if isclass(getattr(module, x))]
                            components = [x for x in classes if (
                                issubclass(getattr(module, x), BaseComponent)
                                or issubclass(getattr(module, x), BasePipeline)
                                or issubclass(getattr(module, x), BaseRunner)
                                )]
                            globals().update({name: getattr(module, name)
                                            for name in components})
                        except: 
                            pass 

    warnings.resetwarnings()


__map_extensions__()



from lolpop.component.data_synthesizer.base_data_synthesizer import BaseDataSynthesizer
from lolpop.utils import common_utils as utils
import pandas as pd 
from sdv.metadata import SingleTableMetadata

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class SDVDataSynthesizer(BaseDataSynthesizer):

    __DEFAULT_CONF__ = {
        "config": {
            "synthesizer" : "SingleTablePreset", 
            "sdv_quality_report_name": "sdv_quality_report.pkl", 
            "sdv_diagnostic_report_name": "sdv_diagnostic_report.pkl",
        }
    }

    def load_data(self, source_file_path, file_type="csv", *args, **kwargs):
        """Loads data for data synthesizer. 

        Args:
            source_file_path (string): path to the file (local)
            file_type (str, optional): Type of file. Defaults to "csv".

        Returns:
            pd.DataFrame: dataframe of the data 
            metadata: sdv metadata of the dataframe
        """
        df = utils.create_df_from_file(source_file_path, **kwargs)
        
        metadata = SingleTableMetadata()
        metadata.detect_from_dataframe(data=df)

        return df, metadata 

    def model_data(self, data, metadata, synthesizer_str=None, *args, **kwargs):
        """Creates a synthesizer of the given data. 

        Args:
            data (pd.Dataframe): Data to synthesize. 
            metadata (dict): Metadata of the data. Should be ouput from load_data. 
            synthesizer_str (str): Name of the synthesizer class to use

        Returns:
            synthesizer: The Synthesizer Model fit on the data
        """
        synthesizer = None 

        if synthesizer_str is None: 
            synthesizer_str = self._get_config("synthesizer")
        synthesizer_cl = self._get_synthesizer_class(synthesizer_str)

        if synthesizer_cl is not None: 
            if synthesizer_str == "SingleTablePreset": 
                kwargs.update({"name": "FAST_ML"}) #required or SingleTablePreset failes
            synthesizer = synthesizer_cl(metadata, **kwargs)
            synthesizer.fit(data)

        return synthesizer

    def sample_data(self, synthesizer, num_rows, *args, **kwargs):
        """sample data from the given synthesizer

        Args:
            synthesizer (Object): The synthesizer. Should be the output from model_data
            num_rows (int): The number of rows to generate

        Returns:
            pd.DataFrame: A dataframe of synthesized data
        """
        data = synthesizer.sample(num_rows=num_rows)

        return data

    def evaluate_data(self, real_data, synthetic_data, metadata, synthesizer_str, output_dir=None, *args, **kwargs):
        """Evaluates the synthetic data

        Args:
            real_data (pd.Dataframe): The real data. 
            synthetic_data (pd.Dataframe): The fake data. 
            metadata (dict): Data's metadata. 

        Returns:
            quality_report: report on the quality of the fake data
            diagnostic_report: diagnostic results about the fake data
        """
        if synthesizer_str is None: 
            synthesizer_str = self._get_config("synthesizer")

        if output_dir is None: 
            output_dir = self._get_config("local_dir")

        evaluator_cl = self._get_evaluator_class(synthesizer_str)
        quality_report = evaluator_cl(real_data = real_data, synthetic_data=synthetic_data, metadata=metadata)
        quality_report.save(filepath="%s/%s" % (output_dir,
                            self._get_config("SDV_QUALITY_REPORT_NAME")))
        
        diagnostic_cl = self._get_diagnostic_class(synthesizer_str)
        diagnostic_report = diagnostic_cl(
            real_data=real_data, synthetic_data=synthetic_data, metadata=metadata)
        diagnostic_report.save(
            filepath="%s/%s" % (output_dir, self._get_config("SDV_DIAGNOSTIC_REPORT_NAME")))
        
        #can also print some charts via sdv.evaluation.single_table.get_column_lot/get_column_pair_plot

        return [quality_report, diagnostic_report]

    def _get_synthesizer_class(self, synthesizer): 
        """
        Loads a synthesizer class based on the given string.

        Args:
            synthesizer (str): String representing the synthesizer.

        Returns:
            synthesizer class: The synthesizer class.

        """
        cl = None
        if synthesizer == "SingleTablePreset":
            cl = utils.load_class(synthesizer, "lite", "sdv")
        elif synthesizer in ["CTGANSynthesizer", "CopulaGANSynthesizer", "GaussianCopulaSynthesizer", "TVAESynthesizer"]:
            cl = utils.load_class(synthesizer, "single_table", "sdv")
        elif synthesizer == "HMASynthesizer": 
            cl = utils.load_class(synthesizer, "multi_table", "sdv")
        elif synthesizer == "PARSynthesizer": 
            cl = utils.load_class(synthesizer, "sequential", "sdv")
        else: 
            msg = "Unsupported synthesizer: %s" % synthesizer
            self.log(msg, level="ERROR")
            raise Exception(msg)
    
        return cl 

    def _get_evaluator_class(self, synthesizer):
        """
        Loads an evaluator class based on the given synthesizer. 

        Args:
            synthesizer (str): String representing the synthesizer. 

        Returns:
            evaluator class: The evaluator class.

        """
        cl = None
        if synthesizer in ["SingleTablePreset", "CTGANSynthesizer", "CopulaGANSynthesizer", "GaussianCopulaSynthesizer", "TVAESynthesizer"]:
            cl = utils.load_class("evaluate_quality",
                                  "single_table", "sdv.evaluation")
        elif synthesizer == "HMASynthesizer":
            cl = utils.load_class("evaluate_quality",
                                  "multi_table", "sdv.evaluation")
        elif synthesizer == "PARSynthesizer":
            cl = utils.load_class("evaluate_quality",
                                  "sequential", "sdv.evaluation")
        else:
            msg = "Unsupported synthesizer: %s" % synthesizer
            self.log(msg, level="ERROR")
            raise Exception(msg)
        
        return cl 

    def _get_diagnostic_class(self, synthesizer):
        """
        Loads a diagnostic class based on the given synthesizer.

        Args:
            synthesizer (str): String representing the synthesizer.

        Returns:
            diagnostic class: The diagnostic class.

        """
        cl = None
        if synthesizer in ["SingleTablePreset", "CTGANSynthesizer", "CopulaGANSynthesizer", "GaussianCopulaSynthesizer", "TVAESynthesizer"]:
            cl = utils.load_class("run_diagnostic", "single_table", "sdv.evaluation")
        elif synthesizer == "HMASynthesizer":
            cl = utils.load_class("evaluate_quality",
                                  "multi_table", "sdv.evaluation")
        elif synthesizer == "PARSynthesizer":
            cl = utils.load_class("evaluate_quality",
                                  "sequential", "sdv.evaluation")
        else:
            msg = "Unsupported synthesizer: %s" % synthesizer
            self.log(msg, level="ERROR")
            raise Exception(msg)

        return cl 


from lolpop.component.data_checker.base_data_checker import BaseDataChecker
from lolpop.utils import common_utils as utils
import stumpy
import numpy as np 
from matplotlib import pyplot as plt 
from matplotlib.patches import Rectangle

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class StumpyMatrixProfiler(BaseDataChecker):
    __REQUIRED_CONF__ = {
        "config": ["local_dir", "model_target"]
    }


    __DEFAULT_CONF__ = {
        "config": {
            "stumpy_analysis_image_name": "STUMPY_DISCORD_ANALYSIS.PNG",
            "stumpy_window_size": [30],
            "stumpy_num_discords": 3,
            }
    }
    

    def check_data(self, data, *args, **kwargs):
        """
        Method for detecting anomalies/novelties in time series data using STUMPY matrix profiling algorithm and plotting the results.

        This method uses STUMPY algorithm for discord discovery in time series data. The algorithm involves sliding a window across
        the time series data and calculating the matrix profile (a distance profile between subsequences). The subsequence corresponding
        to each discovered anomaly/novelty in the time series data will have a distance profile value greater than other subsequence's
        corresponding to a similar or identical anomaly/novelty.

        Parameters:
        data (pandas.DataFrame): Input time series data.
        args (*args): Optional positional arguments to the method.
        kwargs (**kwargs): Optional keyword arguments to the method.

        Returns:
        tuple(None, str, str): Tuple of Nones and the path of the output image, and a string 'ERROR', 'WARN', 'PASS' based on the validity of the input data.
        """
        model_target = self._get_config("MODEL_TARGET")

        windows = self._get_config("stumpy_window_size")
        num_discords = self._get_config("stumpy_num_discords")

        #set up plot
        fig, axs = plt.subplots(len(windows)+1, sharex=True, gridspec_kw={'hspace': 0})
        plt.suptitle('Discord (Anomaly/Novelty) Discovery', fontsize='30')
        axs[0].plot(data[model_target].values)
        axs[0].set_ylabel('TS', fontsize='10')
        box_height = int(data[model_target].max())

        j=1
        for m in windows: 
            #calculate matrix profile
            mp = stumpy.stump(data[model_target], m)
            discords = []
            #find top n discords. These are the most likely anomalies in the data.
            for i in range(1, num_discords+1):
                discord_idx = np.argsort(mp[:, 0])[-i]
                nearest_neighbor_distance = mp[discord_idx, 0]
                discords.append(discord_idx)
                self.log("Found possible anomaly at index %s, which is %s units from its nearest neighbor" %(discord_idx, nearest_neighbor_distance))    
            #plot the matrix profile
            self.__plot_mp(axs, m, box_height, mp, discords, j)
            j=j+1

        #save plot
        file_path = "%s/%s" %(self._get_config("local_dir"), self._get_config("stumpy_analysis_image_name"))
        plt.savefig(file_path)

        return None, file_path, "PASS"

    def __plot_mp(self, axs, m, h, mp, discords, i): 
        """
        Private method for plotting the STUMPY matrix and marking the discovered anomalies/novelties.

        This method plots the STUMPY matrix for the given subsequence length and marks the discovered anomalies/novelties in the time series data.

        Parameters:
        axs (list of matplotlib axes): List of matplotlib axes to plot the matrix.
        m  (int): Subsequence length for STUMPY matrix.
        h (int): Height of box for marking the anomalies/novelties.
        mp (numpy array): STUMPY matrix to be plotted.
        discords (list of int): Indices of the discovered anomalies/novelties in the time series data.
        i (int): Index in axs list where the plot will be made.

        Returns:
        None
        """
        axs[i].set_ylabel('MP (m=%s)' %m, fontsize='10')
        axs[i].set_xlabel('Time', fontsize='10')
        axs[i].plot(mp[:, 0])
         
        for discord_idx in discords: 
            if i == 1:  # only add rectangles for the first window, otherwise it's going to get cluttered.
                rect = Rectangle((discord_idx, 0), m, h, facecolor='lightgrey')
                axs[0].add_patch(rect)
            axs[i].axvline(x=discord_idx, linestyle="dashed")



from lolpop.pipeline.predict.base_predict import BasePredict
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class OfflinePredict(BasePredict): 
    __REQUIRED_CONF__ = {
        "component": ["data_connector", "metadata_tracker", "resource_version_control", "model_explainer", "data_checker", "data_profiler"], 
        "config": []
    }

       
    def compare_data(self, model_version, dataset_version, data, *args, **kwargs):
        """
        Given a trained model version, a dataset version, and a data sample, compares the current 
        data sample with the training data and logs the report of the comparison to the metadata 
        tracker.

        Args:
        model_version(object): The trained model version.
        dataset_version(object): The dataset version.
        data(pandas.DataFrame): A dataframe containing the data to compare.

        Returns:
        None
        """

        #get training dataset version and df 
        vc_info = self.metadata_tracker.get_vc_info(model_version, key="X_train_hexsha")
        train_df = self.resource_version_control.get_data(model_version, vc_info, key = "X_train")

        #compare current dataset version with previous dataset version
        if train_df is not None and not train_df.empty: 
            comparison_report, file_path = self.data_profiler.compare_data(data, train_df)

            self.metadata_tracker.log_data_comparison(
                dataset_version,
                file_path = file_path, 
                report = comparison_report, 
                profiler_class = self.data_profiler.name
                )
        
    def get_predictions(self, model, model_version, data, *args, **kwargs):
        """
        Given a trained model, a model version, and a data sample, returns a pandas dataframe containing 
        predictions and optionally prediction probabilities and explanations, and logs the prediction 
        metrics to the metrics tracker.

        Args:
        model(object): A trained model object.
        model_version(object): The model version.
        data(pandas.DataFrame): A dataframe containing the data to predict from.

        Returns:
        A tuple containing the dataframe of predictions and the unique identifier of the prediction job.
        """

        #get prediction job
        prediction_id = "%s_predictions" %self.metadata_tracker.get_resource_id(model_version)
        prediction_job = self.metadata_tracker.create_resource(id=prediction_id, type="prediction_job", parent=model_version, prediction_count=data.shape[0])

        #drop any metadata columns
        df = data.drop(self._get_config("DROP_COLUMNS", []), axis=1, errors="ignore")

        #make predictions
        transformed_data = model.transform_data(df)
        data["prediction"] = model.predict_df(transformed_data)
        if self.problem_type == "classification": 
            data["prediction_proba"] = model.predict_proba_df(
                transformed_data, to_list=True)

        #get explanations
        if self.problem_type !="timeseries" and not self._get_config("skip_prediction_explanations", False):
            data["explanations"] = self.model_explainer.get_explanations(
                transformed_data, model, model_version, "prediction", to_list=True)

        #log predictions
        self.metrics_tracker.log_prediction_metrics(prediction_job, data["prediction"])

        return data, prediction_job

    def track_predictions(self, prediction_job, data, *args, **kwargs):
        """
        Adds a new data sample as a new version to an existing vc resource.

        Args:
        prediction_job(object): The prediction job.
        data(pandas.DataFrame): A dataframe containing the data to append to the existing vc resource.

        Returns:
        None
        """

        #version data
        vc_info = self.resource_version_control.version_data(prediction_job, data)

        #register version control metadata w/ metadata tracker
        self.metadata_tracker.register_vc_resource(prediction_job, vc_info, file_type="csv")

    def analyze_prediction_drift(self, dataset_version, prediction_job, data, *args, **kwargs):
        """
        Given a dataset version, a prediction job, and a data sample, analyzes drift in the performance 
        of the model between the current dataset and the previous version and logs the report of the 
        analysis to the metadata tracker.

        Args:
        dataset_version(object): The dataset version.
        prediction_job(object): The prediction job.
        data(pandas.DataFrame): A dataframe containing the data to compare.

        Returns:
        None
        """

        #get previous dataset version & dataframe 
        prev_dataset_version = self.metadata_tracker.get_prev_resource_version(dataset_version)
        if prev_dataset_version is not None: 
            vc_info = self.metadata_tracker.get_vc_info(prev_dataset_version)
            prev_data = self.resource_version_control.get_data(prev_dataset_version, vc_info)

            #compare current dataset version with previous dataset version
            if prev_data is not None and not prev_data.empty:
                comparison_report, file_path = self.data_profiler.compare_data(data, prev_data)

                self.metadata_tracker.log_data_comparison(
                    prediction_job,
                    file_path = file_path, 
                    report = comparison_report, 
                    profiler_class = type(self.data_profiler).__name__
                    )
        
    def check_predictions(self, data, prediction_job, *args, **kwargs):
        """
        Given a predicted data sample and a prediction job, performs data checks on the data sample 
        and logs the report of the checks to the metadata tracker.

        Args:
        data(pandas.DataFrame): A dataframe of the predicted data.
        prediction_job(object): The prediction job.

        Returns:
        None
        """
        #run data checks
        data_report, file_path, checks_status = self.data_checker.check_data(data)

        #log data report to metadata tracker
        self.metadata_tracker.log_checks(
            prediction_job,
            file_path = file_path, 
            report = data_report, 
            checker_class = type(self.data_checker).__name__, 
            type = "prediction"
            )

        if checks_status == "ERROR" or checks_status == "WARN": 
            url = self.metadata_tracker.url
            self.notify("Issues found with data checks. Visit %s for more information." %url, checks_status)

    def save_predictions(self, data, table, *args, **kwargs):
        """
        Given a predicted data sample and the name of a target table, saves the data sample to the target 
        table using the data_connector.

        Args:
        data(pandas.DataFrame): A dataframe of the predicted data.
        table(str): A string representing the name of the target table.

        Returns:
        None
        """
        self.data_connector.save_data(data, table)


from lolpop.component.base_component import BaseComponent
from typing import Any


class BaseFeatureTransformer(BaseComponent):
    __REQUIRED_CONF__ = {"component": ["resource_version_control", "metadata_tracker"]
                         }
    
    def fit(self, data, *args, **kwargs) -> Any: 
        pass 

    def transform(self, data, *args, **kwargs) -> Any:
        pass

    def fit_transform(self, data, *args, **kwargs) -> Any: 
        pass 

    def save(self, experiment, *args, **kwargs):
        """
        Saves the feature transformer and associated metadata using the resource version control.

        Args:
            experiment: The experiment associated with the feature transformer.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None.
        """
        transformer_class = self.name
        vc_info = self.resource_version_control.version_feature_transformer(
            experiment, self.transformer, transformer_class=transformer_class)
        experiment_metadata = {
            "feature_transformer_config": self.config,
            "feature_transformer_class": transformer_class
        }
        self.metadata_tracker.register_vc_resource(
            experiment, vc_info, key="feature_transformer", additional_metadata=experiment_metadata)

    def load(self, source, *args, **kwargs) -> Any:
        pass

    def _get_transformer(self) -> Any: 
        return self.transformer
    
    def _set_transformer(self, transformer): 
        self.transformer = transformer




from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class LocalDataConnector(BaseDataConnector):

    __DEFAULT_CONF__ = {
        "config" : {"save_index": True}
    }

    def get_data(self, source_path, *args, **kwargs):
        """
        Reads a CSV, Parquet, or ORC file from a local storage and returns the data as a pandas DataFrame.

        Args:
            source_path (str): The path to the data source file.

        Returns:
            pandas.DataFrame: Returns data as a pandas DataFrame.
        """
        file_type = source_path.split(".")[-1]
        if file_type == "csv":
            data = pd.read_csv(source_path, engine="pyarrow", **kwargs)
        elif file_type == "parquet" or file_type == "pq":
            data = pd.read_parquet(source_path, engine="pyarrow", **kwargs)
        elif file_type == "orc":
            data = pd.read_orc(source_path, engine="pyarrow", **kwargs)
        else: 
            raise Exception("Unsupported file type provided: %s" %source_path)
        self.log("Successfully loaded data from %s into DataFrame." %source_path)

        return data 

    def save_data(self, data, target_path, save_index=None, *args, **kwargs):
        """
        Writes a CSV or Parquet file to a local storage and returns the saved data.

        Args:
            data (pandas.DataFrame): Data to be saved.
            target_path (str): The path to the target file.

        Returns:
            pandas.DataFrame: Returns saved data.
        """
        if save_index is None: 
            save_index = self._get_config("save_index")
        file_type = target_path.split(".")[-1]
        if file_type == "csv":
            data = data.to_csv(target_path, index=save_index, **kwargs)
        elif file_type == "parquet" or file_type == "pq":
            data = data.to_parquet(target_path, index=save_index, **kwargs)
        self.log("Successfully saved data to %s." %target_path)
        return data



from lolpop.pipeline.base_pipeline import BasePipeline
from typing import Any 

class BaseDeploy(BasePipeline): 
    def __init__(self,pipeline_type="deploy", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)

    def promote_model(self, model_version, *args, **kwargs) -> Any: 
        pass 

    def deploy_model(self, promotion, *args, **kwaargs) -> Any: 
        pass 

    def approve_model(self, promotion, *args, **kwargs) -> Any: 
        pass 

    def check_approval(self, promotion, *args, **kwargs) -> bool: 
        pass


from lolpop.component.model_trainer.base_model_trainer import BaseModelTrainer
from lolpop.utils import common_utils as utils
from xgboost import XGBClassifier, XGBRegressor 

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class XGBoostModelTrainer(BaseModelTrainer): 

    mlflow_module = "xgboost"

    def __init__(self, problem_type=None, params=None, *args, **kwargs): 
        if params is None: 
            params = {} 

        #set normal config
        super().__init__(problem_type = problem_type, params=params, *args, **kwargs)
 
        #set up model based on problem_type
        if self.problem_type == "classification": 
            self.model = XGBClassifier(**params)
        elif self.problem_type == "regression": 
            self.model = XGBRegressor(**params)
        else: 
            msg = "Unsupported problem type (%s) in trainer %s" %(problem_type, self.name)
            self.notify(msg)
            raise Exception(msg)


    def fit(self, data, *args, **kwargs): 
        """Fits XGBoost model on given data.

        Args:
            data (dict): Dictionary containing input features X_train, y_train, X_valid, and X_test.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            XGBoost model object.

        """
        self.log("Starting model training with parameters: %s" %str(self.params))
        self.model.fit(data["X_train"], data["y_train"])
        self.log("Finished model training.")
        return self.model  

    def predict(self, data, *args, **kwargs):
        """Performs prediction using the XGBoost model.

        Args:
            data (dict): Dictionary containing input features X_train, y_train, X_valid, and X_test.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: Dictionary containing train, valid, and test predictions and probablities if applicable.

        """
        predictions = {}
        
        data_train = self._order_features(data["X_train"])
        predictions["train"] = self.model.predict(data_train)
        if self.problem_type == "classification": 
            predictions["train_proba"] = self.model.predict_proba(data_train)

        if data.get("X_valid") is not None: 
            data_valid = self._order_features(data["X_valid"])
            predictions["valid"] = self.model.predict(data_valid)
            if self.problem_type == "classification": 
                predictions["valid_proba"] = self.model.predict_proba(data_valid)

        if data.get("X_test") is not None: 
            data_test = self._order_features(data["X_test"])
            predictions["test"] = self.model.predict(data_test)
            if self.problem_type == "classification": 
                predictions["test_proba"] = self.model.predict_proba(data_test)

        #self.predictions = predictions

        return predictions 


    def predict_df(self, df, *args, **kwargs): 
        """Performs prediction using the XGBoost model on a pandas DataFrame.

        Args:
            df (pandas DataFrame): Dataframe containing input features.

        Returns:
            numpy array: Predicted output.

        """
        return self.model.predict(self._order_features(df))

    def predict_proba_df(self,df, to_list=False, *args, **kwargs): 
        """Performs prediction probabilities estimation of the model on a pandas DataFrame.

        Args:
            df (pandas DataFrame): Dataframe containing input features.
            to_list (bool): Whether to convert the probabilities into a list.

        Returns:
            numpy array or list: Predicted probabilities.

        """
        predictions = self.model.predict_proba(df)
        if to_list: 
            predictions = predictions.tolist()
        return predictions
    
    def _get_feature_names(self, *args, **kwargs):
        """Returns a list of feature names. Mainly used to determine the correct order 
        for passing in feature columns when making predictions
        """ 
        return self.model.feature_names_in_
    
    def _order_features(self, data, *args, **kwargs):
        """Orders the features in the provided data

        Args:
            data (object): Data object containing features. 

        Returns:
            data_out: Data object with the correct order of features
        """
        data_out = data
        feature_names = self._get_feature_names()
        if feature_names is not None:
            data_out = data[feature_names]
        return data_out



from lolpop.component.logger.base_logger import BaseLogger
from colorama import init as colorama_init, Fore, Style
from datetime import datetime 
class StdOutLogger(BaseLogger): 
    

    __DEFAULT_CONF__ = {"config": {"log_level": "DEBUG", "use_line_numbers": False}}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        colorama_init()
        
    def log(self, msg, level="INFO", time=None, process_name=None, line_num=None, *args, **kwargs): 
        """
        Logs the message to standard output.

        Args:
            msg (str): The message to be logged.
            level (str): The level of logging for the message.
            time (str, optional): The timestamp for the message (default=None).
            process_name (str, optional): The name of the process (default=None).
            line_num (int, optional): The line number of the message (default=None).
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        Returns:
            None.
        """
        if self._get_level_value(level) <= self._get_level_value(self._get_config("log_level")):
            msg_out = ""
            if not time:
                formatted_level = self._get_level_format(level)
                time = datetime.utcnow().strftime("%Y/%m/%d %H:%M:%S.%f")
            msg_out = msg_out + "%s [%s] " % (time, formatted_level)
            if process_name: 
                msg_out = msg_out + "<%s%s%s" %(Fore.BLUE,process_name,Style.RESET_ALL)
                if line_num and self._get_config("use_line_numbers"): 
                    msg_out = msg_out + \
                        "|%s%s%s> ::: %s" % (Fore.GREEN, line_num, Style.RESET_ALL,  msg)
                else: 
                    msg_out = msg_out + "> ::: %s" %msg
            else: 
                msg_out = msg_out + "::: %s" %msg
            print(msg_out)

    def _get_level_value(self, level): 
        """
        Returns the numeric logging level value associated with the specified string level value.

        Args:
            level (str): The string logging level value.

        Returns:
            The corresponding numeric logging level value.

        Raises:
            None.
        """
        if level == "NONE": 
            return 0 
        elif level == "FATAL":
            return 1 
        elif level == "ERROR": 
            return 2
        elif level == "WARN" or level == "WARNING": 
            return 3
        elif level == "INFO": 
            return 4 
        elif level == "DEBUG": 
            return 5 
        elif level == "TRACE": 
            return 6 
        elif level == "ALL": 
            return 100 


    def _get_level_format(self, level):
        """
        Returns the level formatted with color codes.

        Args:
            level (str): The level of logging.
        Returns:
            The level formatted with color codes according to its level of severity.
        """
        level_num = self._get_level_value(level)
        if level_num < 3:
            return "%s%s%s" % (Fore.RED, level, Style.RESET_ALL)
        elif level_num == 3:
            return "%s%s%s" % (Fore.YELLOW, level, Style.RESET_ALL)
        else: #>4 = INFO or better 
            return level



from lolpop.runner import BaseRunner


class Base{{cookiecutter.RunnerType}}(BaseRunner):
    #Add required configuration here
    __REQUIRED_CONF__ = {
        "config": []
    }

    #Add default configuration here 
    __DEFAULT_CONF__ = {
        "config": {}
    }

    def __init__(self, conf, **kwargs):
        #set normal config
        self.__file_path__ == __file__
        super().__init__(conf, **kwargs)
        #Add any additional class initialization code here.

    # Write your functions here. The Base Class should primarily just implement an interface for 
    # other classes to implement.   
    def my_function(self, required_arg1, required_arg2, *args, **kwargs):
        pass

    def my_other_function(self, required_arg1, required_arg2, *args, **kwargs):
        pass



from lolpop.utils import common_utils as utils
from anytree import AnyNode, RenderTree
from omegaconf import dictconfig, OmegaConf
import os  
from inspect import currentframe

class BaseIntegration: 

    __REQUIRED_CONF__ = {
        "config": []
    }
    __DEFAULT_CONF__ = {
        "config": {}
    }

    __DEFAULT_INT__ = {
        "component": {
            "logger": "StdOutLogger",
            "notifier": "StdOutNotifier",
            "metadata_tracker": "MLFlowMetadataTracker",
        }
    }
    
    suppress_logger = False
    suppress_notifier = False

    def __init__(self, 
                 conf=None, 
                 parent=None, 
                 integration_type=None, 
                 integration_framework=None, 
                 problem_type="Unknown", 
                 skip_config_validation=False, 
                 dependent_integrations=None, 
                 decorators=None,
                 plugin_mods=None, 
                 plugin_paths=None,
                 is_standalone=False,
                 *args, **kwargs):     
        
        if dependent_integrations is None: 
            dependent_integrations = {}
        if decorators is None: 
            decorators = []
        if plugin_mods is None: 
            plugin_mods = []
        if plugin_paths is None: 
            plugin_paths = []

        # Integration name. Mainly going to be used for referring to this during logging, etc. 
        # self.name = class name
        self.name = type(self).__name__

        # self.module = module. 
        # Built in modules should have the structure: 
        # lolpop.<integration_type>.<type>.<module>, such as: 
        # lolpop.component.metadata_tracker.mlflow_metadata_tracker
        self.module = self.__module__
        
        #figure out what type of integration we are? I.E. runner, pipeline, component, etc. 
        self.integration_type = integration_type
        if self.integration_type is None:
            try:
                self.integration_type = self.module.split(".")[-3]
            except:  # noqa: E722
                self.integration_type = "extension"

        #figure out what type of integration type we are. 
        # i.e. if you are a "runner" then we want to know what "runner_type" you are
        try:
            int_type = self.module.split(".")[-2]
        except:  # using some kind of custom class  # noqa: E722
            int_type = self.module
        self.type = int_type
        #setattr(self, "%s_type" %self.integration_type, int_type)

        #set parent object 
        self.parent = parent 
        self.parent_integration_type = None
        if self.parent is not None:
            self.parent_integration_type = parent.integration_type
        self.problem_type = problem_type
        if (self.problem_type == "Unknown" or self.problem_type is None) and self.parent is not None: 
            self.problem_type = self.parent.problem_type

        #get configuration 
        config = utils.get_conf(conf)
        
        #set up integration framework
        self.integration_framework = integration_framework
        if self.integration_framework is None:
            integration_framework = config.get("integration_framework", {})
            if len(integration_framework) > 0: #framework provided via conf
                self.integration_framework = _get_integration_framework_tree(integration_framework)
            elif self.parent is None and not is_standalone: # no parent, so assume we are at the root and get the default
                self.integration_framework = _get_default_integration_framework()
            else: #no framework provided and we have a parent, so not root, assume we are being dynamically created somewhere, so don't try to traverse the framework
                pass 

        #handle config    
        config = utils.resolve_conf_variables(config)
        valid_conf = OmegaConf.create(config).copy() 
        #set the integration config before we start copying parent config in. 
        #This saves the integration config as whatever is passed in + the default config
        self.config = utils.copy_config_into(valid_conf.get("config", {}), self.__DEFAULT_CONF__.get("config", {}))
        #handle all the configuration inheritance
        OmegaConf.update(
            valid_conf, 
            "config", 
            _inherit_config(
                valid_conf.get("config",{}),
                self.__DEFAULT_CONF__.get("config",{}),
                self
            )
        )

        #validate configuration to ensure we have all requirements
        if not skip_config_validation:
            self._validate_conf(valid_conf, dependent_integrations)

        #handle plugins
        if len(plugin_mods) == 0:
            if len(plugin_paths) == 0:
                plugin_paths = self._get_config("plugin_paths", [])
            file_path = None
            if hasattr(self, "__file_path__"):
                file_path = self.__file_path__
            plugin_mods = utils.get_plugin_mods(self, plugin_paths, file_path)
        self.plugin_mods = plugin_mods

        #set up default integrations
        dependent_integrations = utils.set_up_default_integrations(
                                    self, 
                                    valid_conf, 
                                    self.__DEFAULT_INT__,
                                    plugin_mods=plugin_mods,
                                    skip_config_validation=skip_config_validation,
                                    dependent_integrations=dependent_integrations
                                    )


        #set up decorators
        decorators = decorators + utils.set_up_decorators(
            self, valid_conf, plugin_mods=plugin_mods, dependent_integrations=dependent_integrations
            )
        
        #if we're at the root of our framework, we have to apply decorators here,
        # otherwise, things will be applied when they registered below
        if  (self.integration_framework is not None) and (self.integration_framework.is_root): 
            self = utils.apply_decorators(self, decorators, integration_type=self.integration_type)
    
        #set all passed in integrations 
        for integration_type in dependent_integrations.keys():
            for integration in dependent_integrations.get(integration_type).keys(): 
                setattr(self, integration, dependent_integrations.get(integration_type).get(integration))

        #process children
        if self.integration_framework is not None: 
            for child in self.integration_framework.children:
                self.log("Processing node %s in %s integration framework..." %(child.id, self.integration_type))
                processed_integrations = {}

                #id of the node in the intergration framework should be the integration type
                child_integration_type = child.id
                #pop off config so it's not processed as an integration
                integration_config = config.get(child_integration_type, {}).pop("config",{})

                for integration in config.get(child_integration_type, {}).keys():
                    #skip registration if the integration already exists in the default_integrations
                    if integration not in self.__DEFAULT_INT__.get(child_integration_type, {}).keys():
                        self.log("Declared %s integration: %s. Processing..." % (child_integration_type, integration))

                        obj = utils.register_integration_class(self, 
                                                                config, 
                                                                integration,
                                                                integration_type=child_integration_type,
                                                                integration_framework=child,
                                                                problem_type=self.problem_type, 
                                                                dependent_integrations=dependent_integrations,
                                                                plugin_mods=plugin_mods,
                                                                decorators=decorators,
                                                                skip_config_validation=skip_config_validation)
                        if obj is not None: 
                            self.log("Loaded class %s into %s %s" %(type(getattr(self,integration)).__name__, child_integration_type, integration))
                            processed_integrations[integration] = obj
                        else: 
                            self.log("Unable to load class for %s: %s" %(child_integration_type, integration))
                    else: 
                        self.log("Integration %s already in default integrations. Skipping..." %integration)

                ##if integration is a leaf then we need to update other integrations with the full set
                #if child.is_leaf: 
                #    for obj in processed_integrations.values(): 
                #        obj._update_integrations(integrations=processed_integrations)

                #check to see if we want to pass integrations to siblings
                if integration_config.get("pass_integration_to_siblings", True):
                    #add processed integrations to the dependent list so they get passed down.
                    self.log("Updating dependent integrations with all processed child integrations in integration %s. All sibling integrations will have access to processed integrations: %s" %(child.id, str(list(processed_integrations.keys()))), level="DEBUG")
                    if child.id not in dependent_integrations.keys(): 
                        dependent_integrations[child.id] = processed_integrations
                    else: #update existing integration type 
                        dependent_integrations[child.id].update(processed_integrations)
                else: 
                    self.log("Detected that sibling integration pass down is disabled for integration %s. Siblings will not be updated with processed integrations: %s." %(child.id, str(list(processed_integrations.keys()))), level="DEBUG")

                #check to sese if we need to update other integrations with the full set
                # defaults to true if the child is a leaf
                if integration_config.get("update_peer_integrations", child.is_leaf):
                    self.log("Updating all peer integrations for integration %s." % child.id, level="DEBUG")
                    for obj in (list(processed_integrations.values()) + 
                                [v for k,v in dependent_integrations[child.id].items() if k in self.__DEFAULT_INT__.get(child.id,{}).keys()]):
                        obj._update_integrations(integrations=processed_integrations)
                else: 
                    self.log("Detected that peer integration updating is disabled for integration %s. Peer integrations will not be updated." %child.id,level="DEBUG")

    def _update_integrations(self, integrations=None, *args, **kwargs):
        if integrations is None: 
            integrations = {} 
        for integration in integrations.keys():
            setattr(self, integration, integrations.get(integration))

    def _validate_conf(self, conf, dependent_integrations):
        integrations=dependent_integrations.copy()   
        missing, total_missing = utils.validate_conf(conf, self.__REQUIRED_CONF__)
        if total_missing > 0:
            #check to see if missing integrations are passed in
            for integration_type in missing.keys(): 
                missing_found = []
                for integration in missing.get(integration_type): 
                    #make sure to check for class-specific requirements
                    integration_key=integration
                    if "|" in integration_key:
                        integration_key = integration_key.split("|")[0]
                    if integrations.get(integration_type, {}).get(integration_key) is not None:
                        #if missing int is part of passed in dependent ints, then decrement count and remove from missing
                        total_missing = total_missing - 1
                        missing_found.append(integration)
                #need to remove after the above loop so you don't mess up the indices in the middle of a traversal
                for key in missing_found: 
                    missing[integration_type].remove(key)
            if total_missing > 0:
                raise Exception("Missing the following from %s configuration: %s" % (
                    type(self).__name__, missing))

    def log(self, msg, level="INFO", **kwargs): 
        if not self.suppress_logger: 
            self.logger.log(msg, level, process_name=self.name,
                            line_num=currentframe().f_back.f_lineno, **kwargs)

    def notify(self, msg, level="ERROR"): 
        if not self.suppress_notifier: 
            self.notifier.notify(msg, level)
            self.log("Notification Sent: %s" %msg, level)

    #helper function for lookup up config key
    def _get_config(self, key, default_value=None):
        key = key.lower()
        value = utils.lower_conf(self.config).get(key, None)

        #if value wasn't found, check in parent
        if value is None and self.parent is not None: 
            value = self.parent._get_config(key)
        
        #if no value found in parent, check environment
        if value is None: 
            value = os.getenv(key,default=default_value)
        
        return value
        
    def _set_config(self, key, value): 
        key = key.lower()
        self.config[key]=value

    def _print_integration_framework(self): 
        if hasattr(self, "integration_framework"):
            print(RenderTree(self.integration_framework))
        else: 
            print("Object has no integration framework defined.")

    def _print_integrations(self): 
        if hasattr(self, "integration_framework"):
            print("Integrations attached to %s %s:" %(self.integration_type, self.name))
            for child in self.integration_framework.children: 
                integration_type = child.id 
                print("\n %s" %integration_type)
                for integration in [x for x in dir(self) if (not x.startswith("_") and hasattr(getattr(self,x),"integration_type"))]:
                    int_obj = getattr(self, integration)
                    if getattr(int_obj,"integration_type") == integration_type: 
                        print(" -- %s (%s)" %(integration, int_obj.name))
        else:
            print("Object has no integration framework defined.")




def _get_default_integration_framework():
    runner = AnyNode(id="runner")
    global_component = AnyNode(id="component", parent=runner)
    pipeline = AnyNode(id="pipeline", parent=runner)
    component = AnyNode(id="component", parent=pipeline)
    return runner


def _get_integration_framework_tree(framework_conf, parent=None):
    if len(framework_conf) > 1 and parent is None: 
        raise Exception("ERROR: Detected more than one root node in integration framework. You may only have one one root node in your integration framework tree. Found root nodes: %s" %str(framework_conf.keys()))
    node = AnyNode()
    for k, v in framework_conf.items():
        node = AnyNode(id=k)
        if parent is not None:
            node.parent = parent
        if isinstance(v, dictconfig.DictConfig):
            _ = _get_integration_framework_tree(v, parent=node)
    return node


def _inherit_config(conf, default_conf, integration, conf_array=None):

    if conf_array is None: 
        conf_array = []

    parent = integration.parent 

    #parent exists, so we need to iterate
    if parent is not None: 
        conf_array.append(parent.config)
        return _inherit_config(conf, default_conf, parent, conf_array=conf_array)
    else: #no parent, so we can stop
        conf_array.append(conf)
        final_conf = default_conf

        for c in conf_array: 
            final_conf = utils.copy_config_into(c, final_conf)

        return final_conf 


from lolpop.runner.base_runner import BaseRunner
from lolpop.utils import common_utils as utils


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class ClassificationRunner(BaseRunner):

    __REQUIRED_CONF__ = {
        "pipeline": ["process", "train", "deploy", "predict"],
        "component": ["metadata_tracker", "metrics_tracker", "resource_version_control"],
        "config": ["train_data", "eval_data", "prediction_data", "model_target"]
    }

    def __init__(self, problem_type="classification", *args, **kwargs):
        super().__init__(problem_type=problem_type, *args, **kwargs)

    def process_data(self, source = "train", *args, **kwargs):
        """
        Processes data and performs transformations, tracking, profiling, checks, and comparisons between different instances of data.

        Args:
            source (str): Default is "train". Indicates the source of the data. 

        Returns:
            data (pd.DataFrame): DataFrame version of the processed data.
            dataset_version (object): The dataset version object.
        """

        #run data transformations and encodings
        source_data_name = "%s_data" % source
        source_data = self._get_config(source_data_name)
        data = self.process.transform_data(source_data)  # maybe better called get_training_data?

        #track & version data
        dataset_version = self.process.track_data(data, source_data_name)

        #profile data
        self.process.profile_data(data, dataset_version)

        #run data checks
        self.process.check_data(data, dataset_version)

        #run data comparison/drift
        self.process.compare_data(data, dataset_version)

        #return data
        return data, dataset_version

    def train_model(self, data, dataset_version=None, *args, **kwargs):
        """
        Trains the model on the processed data and performs analysis, checks, bias checks, versioning, building lineage, and comparisons between different model versions.

        Args:
            data (pd.DataFrame): DataFrame version of the processed data.
            dataset_version (object): The dataset version object.

        Returns:
            model_version (object): The model version object
            model (Any): The trained model object.
            is_new_model_better (bool): Boolean value of whether or not a better model version was trained.

        Raises:
            Exception: If no data is available to train the model.
        """
        if data is None:
            if dataset_version is not None:
                data = self.resource_version_control.get_data(
                    dataset_version, self.metadata_tracker.get_vc_info(dataset_version))
            else:
                self.notify(
                    "No data provided. Can't train a model", level="ERROR")
                raise Exception("No data provided. Can't train a model.")

        #split data
        data_dict = self.train.split_data(data)

        #train a model
        model, model_version = self.train.train_model(data_dict)

        #analyze the model
        self.train.analyze_model(data_dict, model, model_version)

        #run model checks
        self.train.check_model(data_dict, model, model_version)

        #run bias checks
        self.train.check_model_bias(data_dict, model, model_version)

        #build lineage
        self.metadata_tracker.build_model_lineage(
            model_version, self.process.datasets_used)

        #run comparison to previous model verison
        is_new_model_better = self.train.compare_models(
            data_dict, model, model_version)

        #if new model is better, retrain on all data if specified
        if is_new_model_better:
            if self.train._get_config("retrain_all"):
                model, experiment = self.train.retrain_model_on_all_data(
                    data_dict, model_version, ref_model=model)

        return model_version, model, is_new_model_better

    def deploy_model(self, model_version, model, *args, **kwargs):
        """
        Deploys the trained model if it is approved.

        Args:
            model_version (object): The version of the trained model.
            model (Any): The trained model object.

        Returns:
            deployment (Any): The model deployment object.
        """
        #promote model
        promotion = self.deploy.promote_model(model_version, model)

        #check if model is approved
        is_approved = self.deploy.check_approval(promotion)

        #if approved, deploy model
        if is_approved:
            deployment = self.deploy.deploy_model(promotion, model_version)

        return deployment

    def predict_data(self, model_version, model, data, dataset_version, *args, **kwargs):
        """
        Predicts outcomes for a given dataset using the trained model, and performs data comparison, tracking, drift analysis, and checks.

        Args:
            model_version (object): The version of the trained model.
            model (object): The trained model object.
            data (pd.DataFrame): DataFrame version of the dataset to run predictions on.
            dataset_version (object): The version of the dataset.

        Returns:
            data (pd.DataFrame): DataFrame version of the dataset with predictions and explanations.
            prediction_job (str): The version of the prediction job.
        """
        if model is None: 
            if model_version is not None: 
                experiment = self.metadata_tracker.get_winning_experiment(model_version)
                model_obj = self.resource_version_control.get_model(experiment)
                model = self.metadata_tracker.load_model(model_obj, model_version, model)
        
        #compare evaluation data to training data
        self.predict.compare_data(model_version, dataset_version, data)

        #get predictions
        data, prediction_job = self.predict.get_predictions(model, model_version, data)

        #version data
        self.predict.track_predictions(prediction_job, data)

        #calculate drift
        self.predict.analyze_prediction_drift(dataset_version, prediction_job, data)

        #run prediction checks
        self.predict.check_predictions(data.drop(["explanations", "prediction_proba"],axis=1, errors="ignore"), prediction_job)

        #run save predictions
        self.predict.save_predictions(data, self._get_config("prediction_data"))

        return data, prediction_job

    def evaluate_ground_truth(self, prediction_job=None, *args, **kwargs): 
        """
        Evaluates prediction data against actual data.

        Args:
            prediction_job (object): Default is None. The version of the prediction job.
        """
        #if prediction job isn't known, get the most recent job
        if prediction_job is None: 
            model = self.metadata_tracker.get_resource(self._get_config("model_name"), type="model")
            prediction_job = self.metadata_tracker.get_latest_model_resource(model, type="prediction_job")
        
        #get prediciton data 
        vc_info = self.metadata_tracker.get_vc_info(prediction_job)
        prediction_data = self.resource_version_control.get_data(prediction_job, vc_info) 
        
        #get *current* training data
        train_data = self.process.transform_data(self._get_config("train_data"))
        
        #let's find common index values that exist in both the training dataset and the prediction set
        index = self._get_config("model_index")
        common_indices = train_data.merge(prediction_data, on=index, how="inner")[index]

        #if there are any in common, then we can check the ground truth 
        if len(common_indices) > 0: 
            #filter out any values not in the common index and then sort data so they match indices 
            train_data_filtered = train_data[train_data[index].isin(common_indices)]
            prediction_data_filtered = prediction_data[prediction_data[index].isin(common_indices)]
            train_data_sorted = train_data_filtered.sort_values(by=index).reset_index(drop=True)
            prediction_data_sorted = prediction_data_filtered.sort_values(by=index).reset_index(drop=True)
            ground_truth = {"y_train": train_data_sorted[self._get_config("model_target")]}
            predictions = {"train": prediction_data_sorted["prediction"]}
          
            #get model object and calculate metrics
            model_version = self.metadata_tracker.get_prediction_job_model_version(prediction_job)
            experiment = self.metadata_tracker.get_winning_experiment(model_version)
            model_obj = self.resource_version_control.get_model(experiment)
            model = self.metadata_tracker.load_model(model_obj, model_version, None)
            metrics_val = model.calculate_metrics(ground_truth, predictions, self.train._get_config("metrics"))

            #log stuff
            self.metrics_tracker.log_metrics(prediction_job, metrics_val, self.train._get_config("perf_metric"))
            self.metrics_tracker.log_metric(prediction_job, "num_ground_truth_observations", train_data_sorted.shape[0])
        else: #nothing in prediction has a ground truth yet 
            self.notify("Current training data has no overlap with the prediction job %s" %self.metadata_tracker.get_resource_id(prediction_job), level = "WARNING")

    def stop(self):
        """
        Makes sure that all current metadata is saved once the process is stopped.
        """
        self.metadata_tracker.stop()

    def build_all(self, *args, **kwargs):
        """
        Performs all phases of the classification task including data processing, model training, model deployment, prediction, and evaluation, and notifies the user if a new version of the model is better.
        """
        data, dataset_version = self.process_data()
        model_version, model, is_new_model_better = self.train_model(data, dataset_version)
        if is_new_model_better: 
            deployment = self.deploy_model(model_version, model)
        eval_data, eval_dataset_version = self.process_data(source="eval")
        data, prediction_job = self.predict_data(model_version,model, eval_data, eval_dataset_version)
        #self.evaluate_ground_truth(prediction_job)
        self.stop()
    



from .base_{{cookiecutter.pipeline_type}} import Base{{cookiecutter.PipelineType}}
from lolpop.utils import common_utils as utils
#import your libraries here

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class {{cookiecutter.PipelineClass}}(Base{{cookiecutter.PipelineType}}):
    #Override required or default configurations here for your class
    ##Add required configuration here
    #__REQUIRED_CONF__ = {
    #    "config": []
    #}
    ##Add default configuration here
    #__DEFAULT_CONF__ = {
    #    "config": {}
    #}

    def __init__(self, conf, **kwargs):
        #set normal config
        super().__init__(conf, **kwargs)
        #Add any additional class initialization code here.
        
    # Write your functions here. Implement Base class functions and actually write them! 
    def my_function(self, required_arg1, required_arg2, *args, **kwargs):
        #Your code goes here!
        pass

    def my_other_function(self, required_arg1, required_arg2, *args, **kwargs):
        #Your code goes here!
        pass



from lolpop.component.base_component import BaseComponent
from typing import Any 

class BaseResourceVersionControl(BaseComponent): 

    def version_data(self, dataset_version, data, *args, **kwargs) -> dict[str, Any]: 
        pass 

    def get_data(self, dataset_version,  *args, **kwargs) -> Any :
        pass

    def version_model(self, experiment, model, *args, **kwargs) -> dict[str, Any]:
        pass
    
    def get_model(self, experiment, *args, **kwargs) -> Any : 
        pass

    def version_feature_transformer(self, experiment, transformer, *args, **kwargs) -> dict[str, Any]: 
        pass 

    def get_feature_transformer(self, experiment, model, *args, **kwargs) -> Any: 
        pass 



from lolpop.component.cache_manager.base_cache_manager import BaseCacheManager
from lolpop.utils import common_utils as utils
from pathlib import Path
from inspect import getsource 
import os 
import joblib


@utils.decorate_all_methods([utils.error_handler])
class LocalCacheManager(BaseCacheManager):

    __DEFAULT_CONF__ = {
        "config": {"cache_dir" : "/tmp/cache", 
                   "decorator_method": "cache_decorator",
                   "integration_types": ["component"]}
    }

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)

        self.cache_dir = Path(self._get_config("cache_dir"))
        os.makedirs(self.cache_dir, exist_ok=True)

    def cache(self, key, value, *args, **kwargs):
        """Caches the value for the given key.

        The value can be any object. If the value is a callable, its source code will be stored instead.
        The key determines the filename under which the value will be stored.

        Args:
            key (str): The key to associate with the value.
            value (object): The value to be cached.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            str: The path to the cached file.

        Raises:
            None.
        """
        file = self.cache_dir / key
        if callable(value): 
            value = getsource(value)
        try: 
            joblib.dump(value,file)
            self.log("Wrote file %s to cache." %file, level="DEBUG")
        except Exception as e: 
            self.log("Unable to cache %s to %s: %s. Subsequent workflow runs will have to recompute this value" %(value, file, str(e)), level="WARN")
        return file 

    def retrieve(self, key, *args, **kwargs): 
        """Retrieves the cached object for the given key.

        If the cached file exists, it will be loaded and returned as an object.
        If the cached file doesn't exist, None will be returned.

        Args:
            key (str): The key associated with the cached object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            object: The cached object for the given key, or None if it doesn't exist.

        Raises:
            None.
        """
        out = None
        file = self.cache_dir / key
        try: 
            out = joblib.load(file)
            self.log("Cache hit for file %s" % file, level="DEBUG")
        except: #if file doesn't exist, it's the first time, so just pass   # noqa: E722
            pass 

        return out 
         



from lolpop.component.base_component import BaseComponent
from lolpop.utils import common_utils as utils
import numpy as np 
from typing import Any 

class BaseModelChecker(BaseComponent): 

    __REQUIRED_CONF__ = {"component": ["metrics_tracker"],
                         "config": ["baseline_method", "baseline_value", "perf_metric"]}

    def check_model(self, data, model, *args, **kwargs) -> tuple[Any, str, str]:
        pass 

    def calculate_model_drift(self, data, model, deployed_model, *args, **kwargs) -> tuple[Any, str]:
        pass

    def get_baseline_comparison(self, data, model, model_version, *args, **kwargs) -> tuple[bool, float]:
        """
        Compare the model performance against a baseline.

        Args:
            data: The data for model evaluation.
            model: The trained model object.
            model_version: The version of the model.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            The metric difference between the model and the baseline.

        """

        #TODO: we are already getting predictions elsewhere. 
        #We should just save them and pass in instead of calculating again
        baseline_method = self._get_config("baseline_method")
        baseline_value = self._get_config("baseline_value")
        test_predictions = model.transform_and_predict(data)
        baseline_predictions = self._get_baseline_predictions(data, baseline_method, baseline_value)

        #TODO: model metrics are being recalculated here, should save and load
        perf_metric = self._get_config("perf_metric")
        model_metrics = model.calculate_metrics(data, test_predictions, [perf_metric])
        baseline_metrics = model.calculate_metrics(data, baseline_predictions, [perf_metric])
        for split in baseline_metrics.keys(): 
            self.metrics_tracker.log_metric(model_version, "baseline_%s_%s" %(split, perf_metric), baseline_metrics[split][perf_metric])
        is_baseline_better, metric_diff = self._get_metric_comparison(model_metrics, baseline_metrics, perf_metric)

        self.metrics_tracker.log_metric(model_version, "model_test_performance_over_baseline", metric_diff)

        if is_baseline_better: 
            self.notify("Model performed worse than baseline by %s%%" %metric_diff, level="WARN")

        return is_baseline_better, metric_diff 

    def compare_models(self, data, model, prev_model, model_version, prev_model_version=None, *args, **kwargs) -> bool: 
        """
        Compare the performance of the current model against the previous model.

        Args:
            data: The data for model evaluation.
            model: The current trained model object.
            prev_model: The previous trained model object.
            model_version: The version of the current model.
            prev_model_version: The version of the previous model (default None).
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            bool: True if the current model is better than the previous model, False otherwise.

        """
        #this should only happen during the first training of a model
        if prev_model_version == model_version:
            is_new_model_better = True 
        else: 
            perf_metric = self._get_config("perf_metric")
            is_new_model_better, metric_diff = self._compare_model_performance(model, prev_model, data, perf_metric, model_version)

        return is_new_model_better

    #default implementation of baseline predictions
    def _get_baseline_predictions(self, data, baseline_method, baseline_value, *args, **kwargs) -> dict[str,Any]: 
        """
        Get baseline predictions based on the baseline method and value.

        Args:
            data: The data for baseline comparison.
            baseline_method: The method for baseline comparison.
            baseline_value: The value used in baseline comparison.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the baseline predictions for different data splits.

        """
        has_test = "X_test" in data.keys()
        has_valid = "X_valid" in data.keys()
        baseline_predictions = {}

        #column = user defined column should be used for baseline comparison
        if baseline_method == "column": 
            baseline_predictions["train"] = data["X_train"][baseline_value]
            baseline_predictions["valid"] = data["X_valid"][baseline_value]
            if has_test: 
                baseline_predictions["test"] = data["X_test"][baseline_value]

        elif baseline_method == "value":
            if self.problem_type == "classification":    
                labels = data["y_train"]
                #this is kind of hacky but ... shrug
                labels.index=labels.values
                label_counts = labels.groupby(level=0).count()

                #classification
                if baseline_value == "class_avg": 
                    weights = [x/len(labels) for x in label_counts]
                    baseline_predictions["train"] = np.random.choice(label_counts.index, p=weights, size=len(data["X_train"]))
                    baseline_predictions["valid"] = np.random.choice(label_counts.index, p=weights, size=len(data["X_valid"]))
                    if has_test: 
                        baseline_predictions["test"] = np.random.choice(label_counts.index, p=weights, size=len(data["X_test"]))

                #classification
                elif baseline_value == "class_most_frequent": 
                    most_frequent_class = sorted(label_counts.items(), key=lambda x: x[1], reverse=True)[0][0] #get most frequent class 
                    baseline_predictions["train"] = [most_frequent_class] * len(data["X_train"])
                    baseline_predictions["valid"] = [most_frequent_class] * len(data["X_valid"])
                    if has_test: 
                        baseline_predictions["test"] = [most_frequent_class] * len(data["X_test"])
            
            elif self.problem_type == "regression":
                labels = data["y_train"]

                if baseline_value == "avg": 
                    mean = labels.mean()
                    baseline_predictions["train"] = [mean] * len(data["X_train"])
                    baseline_predictions["valid"] = [mean] * len(data["X_valid"])
                    if has_test: 
                        baseline_predictions["test"] = [mean] * len(data["X_test"])

                if baseline_value == "mode": 
                    mode = labels.mode()[0]
                    baseline_predictions["train"] = [mode] * len(data["X_train"])
                    baseline_predictions["valid"] = [mode] * len(data["X_valid"])
                    if has_test: 
                        baseline_predictions["test"] = [mode] * len(data["X_test"])

                if baseline_value == "max": 
                    max = labels.max()
                    baseline_predictions["train"] = [max] * len(data["X_train"])
                    baseline_predictions["valid"] = [max] * len(data["X_valid"])
                    if has_test: 
                        baseline_predictions["test"] = [max] * len(data["X_test"])

                if baseline_value == "min": 
                    min = labels.min()
                    baseline_predictions["train"] = [min] * len(data["X_train"])
                    baseline_predictions["valid"] = [min] * len(data["X_valid"])
                    if has_test: 
                        baseline_predictions["test"] = [min] * len(data["X_test"])
                        
                if baseline_value == "median": 
                    median = labels.median()
                    baseline_predictions["train"] = [median] * len(data["X_train"])
                    baseline_predictions["valid"] = [median] * len(data["X_valid"])
                    if has_test: 
                        baseline_predictions["test"] = [median] * len(data["X_test"])

            elif self.problem_type == "timeseries":
                #time series . should be in the form 'last_value_N'
                if baseline_value.startswith("last_value"):
                    shift_value = int(baseline_value.split("_")[-1]) 
                    baseline_predictions["train"] = data["y_train"].shift(shift_value)
                    if has_valid: 
                        baseline_predictions["valid"] = data["y_valid"].shift(shift_value)
                    if has_test: 
                        baseline_predictions["test"] = data["y_test"].shift(shift_value)
                        
                elif baseline_value.startswith("lag_mean"):
                    window_size = int(baseline_value.split("_")[-1])
                    baseline_predictions["train"] = data["y_train"].shift(1).rolling(window_size, min_periods=1).mean()
                    if has_valid: 
                        baseline_predictions["valid"] = data["y_valid"].shift(1).rolling(window_size, min_periods=1).mean()
                    if has_test: 
                        baseline_predictions["test"] = data["y_test"].shift(1).rolling(window_size, min_periods=1).mean()

                elif baseline_value.startswith("lag_max"):
                    window_size = int(baseline_value.split("_")[-1])
                    baseline_predictions["train"] = data["y_train"].shift(
                        1).rolling(window_size, min_periods=1).max()
                    if has_valid:
                        baseline_predictions["valid"] = data["y_valid"].shift(
                            1).rolling(window_size, min_periods=1).max()
                    if has_test:
                        baseline_predictions["test"] = data["y_test"].shift(
                            1).rolling(window_size, min_periods=1).max()
                        
                elif baseline_value.startswith("lag_min"):
                    window_size = int(baseline_value.split("_")[-1])
                    baseline_predictions["train"] = data["y_train"].shift(
                        1).rolling(window_size, min_periods=1).min()
                    if has_valid:
                        baseline_predictions["valid"] = data["y_valid"].shift(
                            1).rolling(window_size, min_periods=1).min()
                    if has_test:
                        baseline_predictions["test"] = data["y_test"].shift(
                            1).rolling(window_size, min_periods=1).min()
                        
                elif baseline_value.startswith("lag_median"):
                    window_size = int(baseline_value.split("_")[-1])
                    baseline_predictions["train"] = data["y_train"].shift(
                        1).rolling(window_size, min_periods=1).median()
                    if has_valid:
                        baseline_predictions["valid"] = data["y_valid"].shift(
                            1).rolling(window_size, min_periods=1).median()
                    if has_test:
                        baseline_predictions["test"] = data["y_test"].shift(
                            1).rolling(window_size, min_periods=1).median()
                        
                #first value will be NaN, so set it to the second just to prevent errors. 
                for key in baseline_predictions.keys(): 
                    baseline_predictions[key].iloc[0]=baseline_predictions[key].iloc[1]

        else: #consider other ways to do baseline comparison
            pass

        return baseline_predictions 

    def _get_metric_comparison(self, champion_metric, challenger_metric, perf_metric, *args, **kwargs) -> tuple[bool, float]: 
        """Compares two metric values and determines which is better. 

        Args:
            champion_metric (dict): champion metric
            challenger_metric (dict): challenger metric
            perf_metric (str): The performance metric to use to determine which metric is best

        Returns:
            bool: is the challenger metric better? True or False 
            float: difference between champion and challenger metric
        """
        old_metric = champion_metric.get("test").get(perf_metric)
        new_metric = challenger_metric.get("test").get(perf_metric)
        lower_is_better = not utils.get_metric_direction(perf_metric)

        if lower_is_better: #True = perf_metric is better if lower
            is_challenger_better = (new_metric < old_metric) 
            metric_diff = (new_metric - old_metric) / max(old_metric, 0.00001) * 100
        else: #wants higher perf metric
            is_challenger_better = (new_metric > old_metric)
            metric_diff = (old_metric - new_metric) / max(old_metric, 0.00001) * 100 

        return is_challenger_better, metric_diff

    def _compare_model_performance(self, model, deployed_model, data, perf_metric, current_model_version, *args, **kwargs) -> tuple[bool, float]: 
        """Commpares model performance to deployed model against a static data set. 

        Args:
            model (object): current model
            deployed_model (object): deployed model
            data (dictionary): dicitonary of training/test data
            perf_metric (_type_): performance metric to use to determine which model is better
            current_model_version (_type_): current model verison to log metrics into

        Returns:
            bool: Is the new model better? True or False
            float: difference between metric values
        """
        current_predictions = model.transform_and_predict(data)
        deployed_predictions = deployed_model.transform_and_predict(data)

        current_metrics = model.calculate_metrics(data, current_predictions, [perf_metric])
        deployed_metrics = deployed_model.calculate_metrics(data, deployed_predictions, [perf_metric])

        is_new_model_better, metric_diff = self._get_metric_comparison(deployed_metrics, current_metrics, perf_metric)

        self.metrics_tracker.log_metric(current_model_version, "deployed_model_perf_metric_diff", metric_diff)

        return is_new_model_better, metric_diff


import typer 
from pathlib import Path 
from lolpop.utils import common_utils as utils
from typing import List
import json
from typer.core import TyperGroup


class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls = NaturalOrderGroup, help="Run workflows with lolpop.")


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))

@app.command("workflow", help="Run a workflow.")
def workflow(
    runner_class: str = typer.Argument(..., help="Runner class."),
    config_file: Path = typer.Option(..., "-c", "--config-file", help="Location of runner configuration file."),
    build_method: str = typer.Option("main", "-b", "--build-method",
                                       help="The method in the runner class to execute."),
    build_args: List[str] = typer.Option([], help="List of args to pass into build_method."),
    build_kwargs: str = typer.Option("{}", help="Dict (as a string) of kwargs to pass into build_method"),
    local_file: Path = typer.Option(None, "-l", "--local-file", help="Local file to use to read class definition instead of reading directly from lolpop. Useful when you want to run a modified local class that isn't properly registered. "),
    skip_validation: bool = typer.Option(False, "--skip-validation", help="Skip configuration validation.")
): 

    plugin_mods = []
    if local_file is not None: 
        cl = utils.load_module_from_file(local_file)
        plugin_mods = [cl.__name__]
    typer.secho("Loading class %s" %runner_class, fg="blue")
    runner_cl = utils.load_class(runner_class, plugin_mods=plugin_mods, class_type="runner")
    if runner_cl is None: 
        typer.secho("Failed to properly load class %s. Exiting..." %runner_class, fg="red")
        raise typer.Exit(code=1)
    typer.secho("Loaded %s!" %runner_class, fg="green")

    typer.secho("Initializing class %s with config file %s" %(runner_class, config_file), fg="blue")
    runner = runner_cl(
        conf=config_file, 
        dependent_integrations={},
        skip_config_validation=skip_validation
        )
    typer.secho("Initialized!", fg="green")
    
    if hasattr(runner, build_method): 
        typer.secho("Executing %s.%s with args %s and kwargs %s" %(runner_class, build_method, build_args, build_kwargs), fg="blue")
        func = getattr(runner, build_method)
        func(*build_args, **json.loads(build_kwargs))
        typer.secho("Workflow completed!", fg="green")        
    else: 
        typer.secho("ERROR: Method %s not found in class %s!" %(build_method, runner_class), fg="red")


import typer
from typer.core import TyperGroup
import lolpop.cli.extensions as cli_extensions

class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()


app = typer.Typer(cls=NaturalOrderGroup, help="Run lolpop CLI extensions.")

@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))

#get all extensions
extensions = [x for x in dir(cli_extensions) if not x.startswith("_") and x != "typer"]
for ext in extensions:
    ext_cl = getattr(cli_extensions, ext)
    #look for the typer entrypoint and add that here
    for method in dir(ext_cl): 
        if isinstance(getattr(ext_cl, method), typer.main.Typer):
            name = ext_cl.__name__.split(".")[-1].replace("_","-")
            app.add_typer(getattr(ext_cl, method), name=name)
    



from lolpop.component.base_component import BaseComponent
from typing import Any 

class BaseModelRepository(BaseComponent): 

    def register_model(self, model_version, model, *args, **kwargs) -> Any: 
        pass 

    def promote_model(self, id, *args, **kwargs): 
        pass 

    def approve_model(self, id, *args, **kwargs): 
        pass 

    def check_approva(self, id, *args, **kwargs) -> bool: 
        pass


from lolpop.pipeline.base_pipeline import BasePipeline
from typing import Any 

class BasePredict(BasePipeline): 
    def __init__(self, pipeline_type="predict", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)

    def compare_data(self, model_version, dataset_version, data, *args, **kwargs):
        pass

    def get_predictions(self, model, model_version, data, *args, **kwargs) -> tuple[Any, Any]:
        pass 

    def track_predictions(self, prediction_job, data, *args, **kwargs):
        pass 

    def analyze_prediction_drift(self, dataset_version, prediction_job, data, *args, **kwargs):
        pass 

    def check_predictions(self, data, prediction_job,  *args, **kwargs):
        pass 

    def save_predictions(self, data, target, *args, **kwargs):
        pass 


from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.bigquery_data_connector import BigQueryDataConnector

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class BigQueryDataTransformer(BaseDataTransformer): 


    __REQUIRED_CONF__ = {
        "config": ["GOOGLE_PROJECT", "GOOGLE_DATASET"]
    }

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)
        self.bigquery_config = utils.load_config(
            ["GOOGLE_KEYFILE", "GOOGLE_PROJECT", "GOOGLE_DATASET"], self.config)

    def transform(self, sql, *args, **kwargs):
        """
        Transforms data using BigQueryDataConnector and returns the transformed data.

        Args:
            sql (str): The SQL query to be executed in BigQuery.

        Returns:
            The transformed data returned by the BigQueryDataConnector._load_data method.
        """
        data = BigQueryDataConnector._load_data(sql, self.bigquery_config)
        return data 



from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.duckdb_data_connector import DuckDBDataConnector


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class DuckDBDataTransformer(BaseDataTransformer):


    __REQUIRED_CONF__ = {
        "config": ["duckdb_path"]
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = self._get_config("duckdb_path")

    def transform(self, sql, *args, **kwargs):
        """
        Transform data using DuckDBDataConnector.

        Parameters
        ----------
        sql: str
            The SQL query to transform the data.

        Returns
        -------
        data: pandas.DataFrame
            The transformed data as a Pandas DataFrame.
        """
        data = DuckDBDataConnector._load_data(sql, self.path)
        return data



from lolpop.component.orchestrator.base_orchestrator import BaseOrchestrator
from lolpop.utils import common_utils as utils
from functools import wraps
import os 
import json 
from importlib import import_module

@utils.decorate_all_methods([utils.error_handler])
class PrefectOrchestrator(BaseOrchestrator): 

    __DEFAULT_CONF__ = {
        "config": {
            "decorator_method": "decorator",
            "integration_types": ["runner", "pipeline", "component"], 
            "prefect_profile": "default",
            "prefect_flow_integration": ["runner", "pipeline"], 
            "prefect_task_integration": ["component"],
            "flow_kwargs": {"log_prints": True}, 
            "cache_tasks": True,
        }
    }

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)

        # check to see if env variables are set. 
        # in prefect env variables override profiles and are set in the root context so
        # there is no (non-super hacky) current way to override them programatically
        # but we can at least provide a warning
        prefect_env_variables = [x for x in os.environ.keys() if x.startswith("PREFECT")]
        if len(prefect_env_variables)>0: 
            self.log("Found PREFECT_* configuration in the environment variables: %s. Environment variables override profile settings in Prefect. Consider unsetting these variables if their use is not intended." %str(prefect_env_variables), level="INFO")

        # we need to switch profiles before importing prefect. The reason being that prefect
        # sets the root context when imported to the active profile. The only way to change this
        # programatically is via prefect.context.use_profile and some context manager acrobatics, 
        # which is going to be complicated for lolpop. It's easier to just execute a profile switch
        profile = self._get_config("PREFECT_PROFILE")
        out, _ = utils.execute_cmd(["prefect", "profile", "use", profile], logger=self)

        #if profile is new, then try to create it and apply settings
        if "not found" in out: 
            utils.execute_cmd(
                ["prefect", "profile", "create", profile], logger=self)

            profile_settings = self._get_config("prefect_settings", {})
            for k, v in profile_settings.items():
                utils.execute_cmd(
                    ["prefect", "config", "set", f"{k}={v}"], logger=self)

        self.log("Using prefect profile %s" %profile, level="INFO")

        #now we can import, and we'll also need to add them to globals
        from prefect import flow, task 
        globals()["flow"] = flow 
        globals()["task"] = task 

    def decorator(self, func, cls):
        """
        Decorates a function with Prefect flow or task.

        Args:
            func (object): The function to decorate.
            cls (objcet): The class containing the function.

        Returns:
            The decorated function.
        """
        @wraps(func)
        def prefect_wrapper(*args, **kwargs):
            obj = args[0]
            #wrap in flow
            if obj.integration_type in self._get_config("prefect_flow_integration"):
                flow_kwargs = self._get_config("flow_kwargs", {})
                #note: the integration_framework object creates a "maximum recursion depth exceeded" error in prefect
                # prefect calls prefect._vendors.fastapi.encoders.jsonable_encoder on parameters you pass into your flow
                # for complex stuff this usually errors out and is handled safely, but the anytree object doesn't error
                # and it seems it gets stuck in a child -> parent -> child loop until it breaks. So, we'll just remove it here.
                # this should be safe, as the integration_framework is really only used during __init__ calls and is otherwise
                # a cosmetic feature
                if hasattr(obj, "integration_framework"):
                    delattr(obj, "integration_framework")
                return flow(func, **flow_kwargs)(*args, **kwargs)  # noqa: F821
            #wrap in task
            elif obj.integration_type in self._get_config("prefect_task_integration"):
                try: #skip wrapping things that are sub components. This will cause nested task error in prefect
                    if obj.parent_integration_type == "component": 
                        raise Exception("Detected task in a task. This is not allowed in prefect, so we'll attempt to run the function as normal.")
                    task_kwargs = self._get_config("task_kwargs",{})
                    cache_key_fn = None
                    if self._get_config("cache_tasks"):
                        from prefect.tasks import task_input_hash 
                        cache_key_fn = task_input_hash
                    return task(func, cache_key_fn = cache_key_fn, **task_kwargs)(*args,**kwargs)  # noqa: F821
                except Exception as e: 
                    return func(*args, **kwargs)
            else: #func gets decorated but it shouldn't be, do what's normal
                return func(*args, **kwargs)
        return prefect_wrapper
    
    def package(self, 
                lolpop_class=None,
                lolpop_module=None,
                lolpop_entrypoint="build_all",
                flow_name="prefect_entrypoint",
                package_type="docker",
                base_image="prefecthq/prefect:2-python3.9", 
                prefect_files="prefect_files/",
                copy_files=None,
                lolpop_install_location="'lolpop[cli,prefect,mlflow,xgboost]'",
                run_cmd = None,
                config_file=None,
                flow_kwargs=None,
                dockerfile_path="Dockerfile",
                docker_image_tag=None,
                push_image=False,
                skip_validation=False,
                create_deployment=False,
                work_pool = None, 
                job_variables = None,
                *args, **kwargs):

        """
        Packages the Prefect flow into a Docker image or any other packaging format supported by Prefect.

        Args:
            lolpop_class (str): The name of the lolpop class to package.
            lolpop_module (str): The name of the lolpop module to package.
            lolpop_entrypoint (str): The name of the lolpop entrypoint to package.
            flow_name (str): The name of the Prefect flow.
            package_type (str): The packaging format for the Prefect flow (default: "docker").
            base_image (str): The base Docker image to use for building the Docker image (default: "prefecthq/prefect:2-python3.9").
            prefect_files (str): The folder containing the Prefect files.
            copy_files (list): The files in the base directory to be copied to the Docker image.
            lolpop_install_location (str): The installation location for the lolpop package.
            run_cmd (str): The command to run the at the end of the docker file.
            config_file (str): The path to the config file.
            flow_kwargs (dict): Additional keyword arguments for the Prefect fkiw
            dockerfile_path (str): Path to the dockerfile to use for a custom image when creating a prefect deployment. Defaults to "Dockerfile",
            docker_image_tag (str): Tag to apply to the docker image after creation.
            push_image (bool): Whether to push the image after creation. Defaults to False.
            skip_validation (bool): Whether to skip component validation on the when loading the config file in the entrypoint script.
            create_deployment (bool): Whether to use a proper prefect deploy method. Defaults to False,
            work_pool (str): The workpool name associated with this package.  
            job_variables (str): Additional variables to be passed into the flow deployment to override the base deployment template. 
        """
        if flow_kwargs is None: 
            flow_kwargs = {}
        if copy_files is None: 
            copy_files = [] 
        if job_variables is None: 
            job_variables = {} 

        self.log("Building entrypoint script for prefect flow...")

        if config_file is None: 
            config_file = f"{prefect_files}/dev.yaml"

        if docker_image_tag is None:
            docker_image_tag = utils.get_docker_string(f"lolpop-{lolpop_class}-{lolpop_entrypoint}:latest")
        if ":" not in docker_image_tag:
            docker_image_tag = docker_image_tag + ":latest"

        #Note: for prefect we need to build a wrapper script that exposes the lolpop entrypoint. 
        #The reason for this is that prefect doesn't allow class methods to be entrypoints to a flow.
        #So, the flow entry point must be a just a standalone method. 
        #See https://github.com/PrefectHQ/prefect/discussions/9277 for more discussion on this. 
        entrypoint_script = self._make_entrypoint_script(
            flow_name=flow_name,
            lolpop_class=lolpop_class, 
            lolpop_module=lolpop_module,
            lolpop_entrypoint=lolpop_entrypoint, 
            config_file=config_file,
            prefect_files=prefect_files, 
            flow_kwargs=flow_kwargs, 
            skip_validation=skip_validation,
            create_deployment=create_deployment,
            work_pool=work_pool,
            job_variables=job_variables,
            docker_image_name=docker_image_tag.split(":")[0],
            docker_image_tag=docker_image_tag.split(":")[1],
            dockerfile=dockerfile_path,
            )

        self.log(f"Successfully built entrypoint script {entrypoint_script}")

        #we can now build a dockerfile w/ the correct entrypoint script. 
        #we need this regardless of if we're going to build the docker file now 
        #or later in a flow deploy method. 
        self.log("Building dockerfile...")

        if run_cmd is None:
            run_cmd = f"python {prefect_files}/run.py"

        dockerfile = self._make_dockerfile(
            base_image=base_image,
            prefect_files=prefect_files,
            copy_files=copy_files,
            lolpop_install_location=lolpop_install_location,
            run_cmd=run_cmd,
            dockerfile_path=dockerfile_path)

        self.log(f"Successfully built dockerfile {dockerfile}")
        
        if work_pool is None:

            docker_image_tag = utils.get_docker_string(docker_image_tag)
            self.log(f"Building docker image {docker_image_tag}...")

            docker_image = self._build_docker_image(
                dockerfile_path=dockerfile,
                docker_image_tag=docker_image_tag,
                push_image=push_image,
            )

            self.log(f"Successfully created docker image {docker_image_tag}!")
        else: 
            self.log(f"Detected work pool {work_pool}. Skipping docker build, as prefect will build an image during deployment.")

    def deploy(self, deployment_name, deployment_type="docker", work_pool=None,
               flow_class=None, flow_entrypoint=None,
               docker_image_name = None, k8s_deployment_manifest=None, dockerfile="Dockerfile",
               push_image=False, job_variables=None, secret_name="prefect-secrets", num_replicas=1, 
               prefect_api_key="prefect-api-key", prefect_api_url="prefect-api-url", 
               image_pull_policy="Never", manifest_path="prefect_files/deployment_manifest.yaml",
               namespace="lolpop", worker_image="prefecthq/prefect:2-python3.9-kubernetes", 
               worker_deployment_manifest="prefect_files/worker_deployment_manifest.yaml",
               service_account="default",
               flow_kwargs=None, deployment_kwargs=None, *args, **kwargs):
        """Deploys the prefect flow

        Args:
            deployment_name (str): Name of the deployment to create. 
            deployment_type (str, optional): Type of deploymen to create. Defaults to "docker".
            work_pool (str, optional): Name of the work pool to deploy into. Defaults to None.
            flow_class (str, optional): The flow class to deploy. Only used if work pool is used. Defaults to None.
            flow_entrypoint (str, optional): The flow entrypoint. Only used if work pool is used. Defaults to None.
            docker_image_name (str, optional): The docker image name to use in the deployment, if using a custom image. Defaults to None.
            k8s_deployment_manifest (str, optional): The path to the k8s deployment manifest to use, if deployment_type = kubernetes. Defaults to None.
            dockerfile (str, optional): The path to the dockerfile to use if building a deployment with a custom image. Defaults to "Dockerfile".
            push_image (bool, optional): Whether to push the image after creation. Only used if deploying a custom image. Defaults to False.
            job_variables (dict, optional): Other job variables to pass into the flow deploy method to override the default template. Defaults to {}.
            secret_name (str, optional): Name of the k8s secret holding the prefect secrets. Only used if deploying into k8s. Defaults to "prefect-secrets".
            num_replicas (int, optional): Number of replica pods to deploy. Only used if deploying into k8s. Defaults to 1.
            prefect_api_key (str, optional): The name of the prefect api key secret in secret_name. Defaults to "prefect-api-key".
            prefect_api_url (str, optional): The name of the prefect api url secret in secret name. Defaults to "prefect-api-url".
            image_pull_policy (str, optional): The image pull policy to use in the k8s deployment manifest.. Defaults to "Never".
            manifest_path (str, optional): Path to a k8s deployment manifest. Only used if deploying into k8s and if you wish to use a custom manifest file. Defaults to "prefect_files/deployment_manifest.yaml".
            namespace (str, optional): k8s namespace to deploy into. Onkly used if deploying into k8s. Defaults to "lolpop".
            worker_image (str, optional): The image to use for the prefect k8s worker. Defaults to "prefecthq/prefect:2-python3.9-kubernetes".
            worker_deployment_manifest (str, optional): The manifest to use for the prefect k8s worker. Defaults to "prefect_files/worker_deployment_manifest.yaml".
            flow_kwargs (dict, optional): kwargs to pass into the flow instatiation. Defaults to {}.
            deployment_kwargs (dict, optional): kwargs to pass into the deployment instatiation. Defaults to {}.

        Raises:
            Nothing
        """
        if flow_kwargs is None: 
            flow_kwargs = {}
        if deployment_kwargs is None: 
            deployment_kwargs = {}
        if job_variables is None: 
            job_variables = {}         

        if docker_image_name is not None:
            if ":" not in docker_image_name:
                docker_image_name = docker_image_name + ":latest"

        #if we have a workpool or deploying local then we'll use the prefect deployment mechanism
        if deployment_type=="local" or work_pool is not None: 
            if flow_class is None and flow_entrypoint is None: 
                raise Exception("flow_class and flow_entrypoint are None. They must be populated for deployment_type = 'local'.")

            from prefect import flow, Flow
            from prefect.deployments import DeploymentImage

            mod = import_module(flow_class)
            f = getattr(mod,flow_entrypoint)
            if not isinstance(f, Flow):
                f = flow(f, **flow_kwargs)
        
            if deployment_type == "local":
                f.serve(name=deployment_name, **deployment_kwargs)
            elif deployment_type == "docker": 
                if docker_image_name is not None: 
                    image_name, image_tag = docker_image_name.split(":")
                    f.deploy(name=deployment_name, work_pool_name=work_pool,  
                             image=DeploymentImage(name=image_name, tag=image_tag, dockerfile=dockerfile),
                             push=push_image, job_variables=job_variables, **deployment_kwargs)
                else: 
                    f.deploy(name=deployment_name, work_pool_name=work_pool, **deployment_kwargs)
            elif deployment_type == "kubernetes": 
                #make sure secrets are applied to k8s
                self._apply_secrets(secret_name, {
                    prefect_api_url: os.environ["PREFECT_API_URL"],
                    prefect_api_key: os.environ["PREFECT_API_KEY"]
                })

                if k8s_deployment_manifest is None:
                    #make deployment script for prefect worker
                    self.log("Building k8s deployment manifest for prefect worker...")

                    k8s_deployment_manifest = self._make_worker_deployment_manifest(
                        work_pool=work_pool, 
                        deployment_name=deployment_name,
                        namespace=namespace,
                        image=worker_image,
                        manifest_path=worker_deployment_manifest,
                        num_replicas=num_replicas, 
                        image_pull_policy=image_pull_policy,
                        prefect_secret_name=secret_name, 
                        key_prefect_api_key = prefect_api_key, 
                        key_prefect_api_url = prefect_api_url,
                        service_account=service_account,
                        )
                    self.log("Built k8s deployment manifest for prefect worker. Saved to %s." %k8s_deployment_manifest)
                    
                #apply deployment
                out, exit_code = utils.execute_cmd(["kubectl", "apply", '-f', k8s_deployment_manifest], logger=self)

                if exit_code==0: 
                    self.log(out)
                else: 
                    self.log("Applying the manifest %s failed. Please inspect the yaml file to determine any errors." %k8s_deployment_manifest)

                #deploy flow 
                if docker_image_name is not None: 
                    image_name, image_tag = docker_image_name.split(":")
                    f.deploy(name=deployment_name, work_pool_name=work_pool,  
                             image=DeploymentImage(name=image_name, tag=image_tag, dockerfile=dockerfile),
                             push=push_image, job_variables=job_variables, **deployment_kwargs)
                else: 
                    f.deploy(name=deployment_name, work_pool_name=work_pool, **deployment_kwargs)

        elif deployment_type=="docker": #no workpool, so we want to just serve the container in docker
            if docker_image_name is None: 
                raise Exception("docker_image_name is None. It must be provided for deployment_type=docker.")
            
            PREFECT_API_URL = f'PREFECT_API_URL={os.environ["PREFECT_API_URL"]}'
            PREFECT_API_KEY = f'PREFECT_API_KEY={os.environ["PREFECT_API_KEY"]}'
            utils.execute_cmd(["docker", "run", '-e', PREFECT_API_URL,'-e', PREFECT_API_KEY, docker_image_name], run_background=True, logger=self)

        elif deployment_type=="kubernetes": #no workpool, so we want to serve a container in k8s deployment
            if k8s_deployment_manifest is None: 
                self.log("Building k8s deployment manifest...")

                self._apply_secrets(secret_name, {
                                    prefect_api_url: os.environ["PREFECT_API_URL"], 
                                    prefect_api_key: os.environ["PREFECT_API_KEY"]
                                    })
                k8s_deployment_manifest = self._make_deployment_manifest(
                    flow_name=utils.get_docker_string(flow_entrypoint), 
                    deployment_name=deployment_name,
                    image=docker_image_name,
                    manifest_path=manifest_path,
                    num_replicas=num_replicas, 
                    image_pull_policy=image_pull_policy,
                    prefect_secret_name=secret_name, 
                    key_prefect_api_key = prefect_api_key, 
                    key_prefect_api_url = prefect_api_url,
                    )
                self.log("Built k8s deployment manifest. Saved to %s." %k8s_deployment_manifest)

            out, exit_code = utils.execute_cmd(["kubectl", "apply", '-f', k8s_deployment_manifest], logger=self)

            if exit_code==0: 
                self.log(out)
            else: 
                self.log("Applying the manifest %s failed. Please inspect the yaml file to determine any errors." %k8s_deployment_manifest)

        else: 
            raise Exception("Deployment type %s not supported for PrefectOrchestrator" %deployment_type)     


    def run(self, deployment_name, *args, **kwargs): 
        """Runs a prefect deployment. 

        Args:
            deployment_name (str): The deployment name to run.

        Returns:
            (flow_id, url): A tuple containing the flow id and the url of the flow run. 
        """

        if kwargs is not None and len(kwargs) > 0: 
            out, code = utils.execute_cmd(["prefect", "deployment", "run", deployment_name, "--params", json.dumps(kwargs)], logger=self)
        else: 
            out, code = utils.execute_cmd(["prefect", "deployment", "run", deployment_name], logger=self)
        
        try:
            flow_id = out.split("UUID:")[1].split("\n")[0].strip()
        except:
            flow_id = out 

        try: 
            url = out.split("URL:")[1].replace("\n","").strip()
        except: 
            url = ""
        return (flow_id, url) 


    def stop(self, deployment_name, deployment_type="docker", docker_image_name = None, *args, **kwargs): 
        """Shuts down a current prefect deployment

        Args:
            deployment_name (str): The deployment name to shut down
            deployment_type (str, optional): The type of deployment that it is. Defaults to "docker".
            docker_image_name (str, optional): The image name used in the deployment. Defaults to None.

        Returns:
            None
        """
        if deployment_type == "docker": 
            if docker_image_name is None: 
                raise Exception("docker_image_name not provided. This is required to stop the running docker container.")

            container_id, code = utils.execute_cmd(["docker", "ps", "-q", "--filter", "ancestor=%s" %docker_image_name], logger=self)

            if code !=0: 
                raise Exception("Error when trying to obtain container id.")
            
            if container_id is None or len(container_id) == 0: 
                self.log("Unable to find running container with image %s. The deployment was possibly already shut down?" %docker_image_name)
                return 
            
            container_id = container_id.strip()
            _, code = utils.execute_cmd(["docker", "stop", container_id], logger=self)

            if code == 0: 
                self.log("Successfully shut down container %s" %container_id)


        elif deployment_type == "kubernetes": 

            _, code = utils.execute_cmd(["kubectl", "delete", "deployment", deployment_name], logger=self)
            
            if code == 0:
                self.log("Successfully deleted deployment %s" % deployment_name)

        elif deployment_type == "local": 
            self.log("Local deployment consumes your terminal. Please find the window serving the prefect deployment and pushg Ctrl+C.") 

        else: 
            raise Exception("Deployment type %s unsupported." %deployment_type)


    # creates an entrypoint script to use in the docker image. 
    # this file gets saved into prefect_files/run.py
    def _make_entrypoint_script(self,
                                flow_name="prefect_entrypoint",
                                lolpop_module = "lolpop.runner",
                                lolpop_class = None,
                                lolpop_entrypoint = "build_all",
                                config_file="prefect_files/dev.yaml",
                                flow_kwargs=None, 
                                prefect_files = "prefect_files",
                                create_deployment=False,
                                skip_validation=False,
                                work_pool=None, 
                                job_variables=None, 
                                docker_image_name=None, 
                                docker_image_tag="latest", 
                                dockerfile="Dockerfile",
                                ): 
        if flow_kwargs is None: 
            flow_kwargs = {} 
        if job_variables is None: 
            job_variables = {}

        if lolpop_class is None: 
            raise Exception("lolpop_class must not be None!")

        flow_kwargs_str = ""
        if len(flow_kwargs)>0: 
            for k,v in flow_kwargs: 
                flow_kwargs_str = flow_kwargs_str + f"{k}={v}, "

        import_text = f"import {lolpop_class}"
        if lolpop_module is not None: 
            import_text = f"from {lolpop_module} {import_text}"
            
        entrypoint = f"{flow_name}()"
        entrypoint_name = utils.get_docker_string(f'lolpop-{lolpop_class.lower()}-{lolpop_entrypoint}')
        if create_deployment: 
            entrypoint = f"{flow_name}.serve(name='{entrypoint_name}')"
            #if work_pool is not None: 
            #    entrypoint = f"{flow_name}.deploy(name='{entrypoint_name}', \
            #                    work_pool_name='{work_pool}', job_variables={job_variables}, \
            #                    image=DeploymentImage(name='{docker_image_name}', tag='{docker_image_tag}', dockerfile='{dockerfile}'),\
            #                    push=False)"
            
        entrypoint_script = f"""{import_text}
from prefect import flow 
from prefect.deployments import DeploymentImage 

@flow({flow_kwargs_str})
def {flow_name}():
    runner = {lolpop_class}(conf="{config_file}", skip_config_validation={skip_validation})
    runner.{lolpop_entrypoint}()
    runner.log("Workflow completed!")

if __name__ == "__main__":
    {entrypoint}
"""
        entrypoint_script_path = f"{prefect_files}/run.py"
        with open(entrypoint_script_path, "w+") as file:
            file.write(entrypoint_script)

        return entrypoint_script_path

    #makes a dockerfile that sets up lolpop and executes the entrypoint script
    def _make_dockerfile(self, 
                          base_image="prefecthq/prefect:2-python3.9",
                          prefect_files="prefect_files/",
                          copy_files=None,
                          lolpop_install_location="'lolpop[cli,prefect,mlflow,xgboost]'",
                          run_cmd="lolpop prefect_files/run.py",
                          dockerfile_path = "Dockerfile"):
        if copy_files is not None and len(copy_files) > 0:
            copy_files = "COPY %s ./" % " ".join(copy_files)
        else: 
            copy_files = ""

        dockerfile = f"""FROM {base_image}

WORKDIR /opt/prefect  
RUN mkdir prefect_files 
COPY {prefect_files} prefect_files/
{copy_files} 
RUN pip install {lolpop_install_location}

# Run our flow script when the container starts
ENV PYTHONPATH=$PYTHONPATH:/opt/prefect
CMD {json.dumps(run_cmd.split(" "))}
        """
        #write dockerfile
        with open(dockerfile_path, "w+") as file: 
            file.write(dockerfile)

        return dockerfile_path

    #create a deployment manifest for the lolpop docker image. 
    #This manifest will deploy the lolpop docker image into k8s
    def _make_deployment_manifest(self,
                                      manifest_path="prefect_files/deployment_manifest.yaml",
                                      deployment_name="lolpop-prefect", 
                                      namespace="lolpop",
                                      num_replicas=1,
                                      flow_name=None,
                                      image=None,
                                      image_pull_policy="Never",
                                      prefect_secret_name="prefect-secrets", 
                                      key_prefect_api_url="prefect-api-url", 
                                      key_prefect_api_key="prefect-api-key",
                                      ): 
        
        if flow_name is None or image is None: 
            raise Exception("flow_name and image are None. These must be provided in order to createa  manifest file.")

        deployment_manifest = f"""apiVersion: apps/v1
kind: Deployment
metadata:
  name: {deployment_name}
  namespace={namespace}
spec:
  replicas: {num_replicas}
  selector:
    matchLabels:
      flow: {flow_name}
  template:
    metadata:
      labels:
        flow: {flow_name}
    spec:
      containers:
      - name: flow-container
        image: {image}
        env:
        - name: PREFECT_API_URL
          valueFrom: 
            secretKeyRef: 
              name: {prefect_secret_name} 
              key: {key_prefect_api_url}
        - name: PREFECT_API_KEY
          valueFrom: 
            secretKeyRef: 
              name: {prefect_secret_name} 
              key: {key_prefect_api_key}
        imagePullPolicy: {image_pull_policy}
"""
        #write manifest
        with open(manifest_path, "w+") as file:
            file.write(deployment_manifest)

        return manifest_path

    #Create a manifest for the prefect worker. 
    #This worker will pool the prefect server for work and run flows. 
    def _make_worker_deployment_manifest(self,
                                         manifest_path="prefect_files/worker_deployment_manifest.yaml",
                                         deployment_name="lolpop-prefect-worker",
                                         namespace="lolpop", 
                                         num_replicas=1,
                                         work_pool=None,
                                         image="prefecthq/prefect:2-python3.9-kubernetes",
                                         image_pull_policy="IfNotPresent",
                                         prefect_secret_name="prefect-secrets",
                                         key_prefect_api_url="prefect-api-url",
                                         key_prefect_api_key="prefect-api-key",
                                         service_account="default",
                                         ):
        if work_pool is None:
            raise Exception(
                "work_pool is None. This must be provided in order to createa  manifest file.")

        deployment_manifest = f"""apiVersion: apps/v1
kind: Deployment
metadata:
  name: {deployment_name}-{work_pool}
  namespace: {namespace}
spec:
  replicas: {num_replicas}
  selector:
    matchLabels:
      work_pool: {work_pool}
  template:
    metadata:
      labels:
        work_pool: {work_pool}
    spec:
      serviceAccountName: {service_account}
      containers:
      - name: prefect-worker
        image: {image}
        command: 
          - /usr/bin/tini
          - -g
          - -- 
          - /opt/prefect/entrypoint.sh
        args:
          - prefect
          - worker
          - start
          - --type
          - kubernetes
          - --pool 
          - {work_pool}
        env:
        - name: PREFECT_API_URL
          valueFrom: 
            secretKeyRef: 
              name: {prefect_secret_name} 
              key: {key_prefect_api_url}
        - name: PREFECT_API_KEY
          valueFrom: 
            secretKeyRef: 
              name: {prefect_secret_name} 
              key: {key_prefect_api_key}
        imagePullPolicy: IfNotPresent
"""
        #write manifest
        with open(manifest_path, "w+") as file:
            file.write(deployment_manifest)

        return manifest_path

    #apply prefect secrets to k8s
    def _apply_secrets(self, secret_name, secret_dict, namespace="lolpop"):
        
        secret_arr = ["kubectl", "create", "secret", "generic", secret_name, "--namespace", namespace]
        for k,v in secret_dict.items(): 
            secret_arr.append(f"--from-literal={k}={v}")
        utils.execute_cmd(["kubectl", "delete", "secret", secret_name,
                          "--ignore-not-found", "--namespace", namespace], logger=self)
        return utils.execute_cmd(secret_arr)

    #build the docker image
    def _build_docker_image(self,
            dockerfile_path="Dockerfile",
            docker_image_tag=None,
            push_image=False,
            docker_kwargs=None
    ): 
        if docker_kwargs is None: 
            docker_kwargs = {}
        if docker_image_tag is None: 
            raise Exception("docker_image_tag is None. Must provided a proper tag for the image.")
        
        if ":" not in docker_image_tag:
            docker_image_tag = docker_image_tag + ":latest"

        utils.execute_cmd(["docker", "build", "-t", docker_image_tag, "-f", dockerfile_path, "."],logger=self)

        if push_image: 
            utils.execute_cmd(["docker", "push", docker_image_tag], logger=self)



import typer 
import inspect 
import os
import json 

from typing import List
from cookiecutter.main import cookiecutter
from pathlib import Path
from typer.core import TyperGroup

from lolpop import __template_path__ as lolpop_template_path
from lolpop.utils import common_utils as utils


class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls = NaturalOrderGroup, help="Create new runners, pipelines, and components.")

LOLPOP_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
#PARENT_DIR = os.path.dirname(LOLPOP_DIR)


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))


@app.command("component", help="Initialize a custom component.")
def component(
    component_type: str = typer.Argument(..., help="Component type (Should be snake_case)."),
    component_class: str = typer.Argument(..., help="Component class name (Should be snake_case)."),
    extension_name: str = typer.Argument(..., help = "Name of the parent extension for this resource."),
    template_path: str = typer.Option(
        lolpop_template_path + "/component_template", help="Path to the template file. Or a git url of a template file."),
    project_dir: Path = typer.Option(os.getcwd(), help="Project directory for the new component."),
): 
    create_template("component", component_type,
                    component_class, template_path, project_dir, extension_name)
    

@app.command("pipeline", help="Initialize a custom pipeline.")
def pipeline(
    pipeline_type: str = typer.Argument(..., help="Pipeline type (Should be snake_case)."),
    pipeline_class: str = typer.Argument(
        ..., help="Pipeline class name (Should be snake_case)."),
    extension_name: str = typer.Argument(
            ..., help="Name of the parent extension for this resource."),
    template_path: str = typer.Option(
        lolpop_template_path + "/pipeline_template", help="Path to the template file. Or a git url of a template file."),
    project_dir: Path = typer.Option(os.getcwd(), help="Parent directory for the new pipeline."),
): 
    create_template("pipeline", pipeline_type,
                    pipeline_class, template_path, project_dir, extension_name)


@app.command("runner", help="Initialize a custom runner.")
def runner(
    runner_type: str = typer.Argument(..., help="Component type (Should be snake_case)"),
    runner_class: str = typer.Argument(..., help="Component class name (Should be snake_case)."),
    extension_name: str = typer.Argument(..., help="Name of the parent extension for this resource."),
    template_path: str = typer.Option(
        lolpop_template_path + "/runner_template", help="Path to the template file. Or a git url of a template file."),
    project_dir: Path = typer.Option(os.getcwd(), help="Parent directory for the new component."),
): 
    create_template("runner", runner_type,
                    runner_class, template_path, project_dir, extension_name)
    

@app.command("cli-command", help="Initialize a custom cli command.")
def cli_command(
    command_name: str = typer.Argument(
        ..., help="Name of the cli command to create. (Use Snake Case)"),
    template_path: str = typer.Option(
        lolpop_template_path + "/cli_template", help="Path to the template file. Or a git url of a template file."),
    command_description: str = typer.Option(None, help="description for the command_name"),
    project_dir: Path = typer.Option(
        os.getcwd(), help="Parent directory for the new component."),
):
    typer.secho("Creating template for %s" %command_name, fg="blue")
    try:
        output_dir = "%s/lolpop/cli/extensions/" % (project_dir)
        cookiecutter(template=template_path,
                     extra_context={
                         "cli_command": command_name,
                         "command_description": command_description,
                     },
                     output_dir=output_dir, no_input=True)
        typer.secho("Successfully created template at location %s" %
                    output_dir, fg="green")
    except Exception as e:
        typer.secho("Failed to create template for %s: %s" %
                    (command_name, str(e)), fg="red")

    
#template type = component, pipeline, runner, 
def create_template(template_type, object_type, object_class, template_path, project_dir, extension_name): 
    typer.secho("Creating template for %s %s" %(template_type, object_class), fg="blue")
    try: 
        object_class_camel = ''.join([i.title() for i in object_class.split("_")])
        object_type_camel = ''.join([i.title() for i in object_type.split("_")])
        #TODO: Check if object_type already exists and just copy a new template in that 
        #directory instead of trying to create a new folder
        output_dir = "%s/lolpop/extension/%s/%s" %(project_dir, extension_name, template_type)
        cookiecutter(template = template_path,
                    extra_context={
                        "%sClass" %template_type.title(): object_class_camel,
                        "%s_class" %template_type: object_class,
                        "%s_type" %template_type: object_type, 
                        "%sType" %template_type.title(): object_type_camel,
                        },
                    output_dir=output_dir, no_input=True)
        typer.secho("Successfully created template at location %s" %
                    output_dir, fg="green")
    except Exception as e: 
        typer.secho("Failed to create template for %s %s: %s" %(template_type, object_class, str(e)), fg="red") 


@app.command("documentation", help="Create documentation for a class.")
def create_documentation(
    source_file: Path = typer.Argument(
        ..., help="Path to the source file."),
    class_name: str = typer.Argument(
        ..., help="Class name in source_file to document."),
    method_filter: List[str] = typer.Option([], "--method-filter", "-f", help="Methods to include in the documentation. default=all."),
    generator_class: str = typer.Option(
        "OpenAIChatbot", "--generator-class", "-g", help="Generative AI Chatbot class name."),
    generator_kwargs: str = typer.Option(
        "{}", "--kwargs", "-k", help="Keyword arguments to pass into the generator class"),
    output_path: Path = typer.Option(None, "--output-path", "-o", help="The location to save the documentation"),
    documentation_format: str = typer.Option(
        "markdown", "--documentation-format", "-d", help="The format you would like the documentation to be written in.")
):
    chatbot = load_chatbot(generator_class, **json.loads(generator_kwargs))
    class_code, num_lines = get_source_from_file(source_file, class_name)

    if num_lines > 0: 
        typer.secho("Preparing prompt for chatbot...", fg="blue")
        messages=[]
        messages.append(chatbot.prepare_message(role="system", content="You are a technical writer for a software company."))
        filter_text = "this class"
        if len(method_filter)>0: 
            filter_text = "these methods in the class: %s" %str(method_filter)
        messages.append(chatbot.prepare_message(role="user", content="The following is a python class. I would like you to write technical documentation for %s. Include a description of each method in the class, as well as providing at least one example of using the class in a small code snippet. Please write your documentation in %s format. Class Code: %s" %(filter_text, documentation_format, class_code)))
        typer.secho("Prompting chatbot and awaiting a response ...", fg="blue")
        response = chatbot.ask(messages=messages, **json.loads(generator_kwargs))
        typer.secho("Successfully queried chatbot. Received response: \n\n %s" %response, fg="green")

        if output_path: 
            with open(output_path, "w+") as f: 
                f.write(response)
            typer.secho("Successfully wrote response out to file %s" %output_path)

    else: 
        typer.secho("No lines found for class %s, aborting. Please check your inputs and try again." %class_name)


@app.command("tests", help="Create tests for a class.")
def create_tests(
    source_file: Path = typer.Argument(
        ..., help="Path to the source file."),
    class_name: str = typer.Argument(
        ..., help="Class name in source_file to document."),
    method_filter: List[str] = typer.Option(
        [], "--method-filter", "-f", help="Methods to include in the documentation. default=all."),
    generator_class: str = typer.Option(
        "OpenAIChatbot", "--generator-class", "-g", help="Generative AI Chatbot class name."),
    generator_kwargs: str = typer.Option(
        "{}", "--kwargs", "-k", help="Keyword arguments to pass into the generator class"),
    output_path: Path = typer.Option(
        None, "--output-path", "-o", help="The location to save the documentation"),
    testing_framework: str = typer.Option("pytest", "--testing-framework", "-t", help="The testing framework you would like the tests to be written in.")
):
    chatbot = load_chatbot(generator_class, **json.loads(generator_kwargs))
    class_code, num_lines = get_source_from_file(source_file, class_name)
    
    if num_lines > 0:
        typer.secho("Preparing prompt for chatbot...", fg="blue")
        messages = []
        messages.append(chatbot.prepare_message(
            role="system", content="You are an expert software developer for a software company."))
        filter_text = "all methods in this class"
        if len(method_filter) > 0:
            filter_text = "these methods in the class: %s" % str(
                method_filter)
        messages.append(chatbot.prepare_message(role="user", content="The following is a python class. I would like you to write tests for %s. Use the %s testing framework and python to write your tests. Please write at least two tests for each method and respond in the form of a python file. Only include the python file and no other text in your response. Class Code: %s" % (filter_text, testing_framework, class_code)))
        typer.secho("Prompting chatbot and awaiting a response ...", fg="blue")
        response = chatbot.ask(messages=messages, **
                               json.loads(generator_kwargs))
        typer.secho(
            "Successfully queried chatbot. Received response: \n\n %s" % response, fg="green")

        if output_path:
            with open(output_path, "w+") as f:
                f.write(response)
            typer.secho("Successfully wrote response out to file %s" %
                        output_path)

    else:
        typer.secho(
            "No lines found for class %s, aborting. Please check your inputs and try again." % class_name)


@app.command("docstrings", help="Create docstrings for a class.")
def create_docstrings(
    source_file: Path = typer.Argument(
        ..., help="Path to the source file."),
    class_name: str = typer.Argument(
        ..., help="Class name in source_file to document."),
    method_filter: List[str] = typer.Option(
        [], "--method-filter", "-f", help="Methods to include in the documentation. default=all."),
    generator_class: str = typer.Option(
        "OpenAIChatbot", "--generator-class", "-g", help="Generative AI Chatbot class name."),
    generator_kwargs: str = typer.Option(
        "{}", "--kwargs", "-k", help="Keyword arguments to pass into the generator class"),
    output_path: Path = typer.Option(
        None, "--output-path", "-o", help="The location to save the documentation"),
    docstring_format: str = typer.Option(
        "Google", "--docstring-format", "-d", help="The format you would like the docstring to be written in.")

    ):
    chatbot = load_chatbot(generator_class, **json.loads(generator_kwargs))
    class_code, num_lines = get_source_from_file(source_file, class_name)

    if num_lines > 0:
        typer.secho("Preparing prompt for chatbot...", fg="blue")
        messages = []
        messages.append(chatbot.prepare_message(
            role="system", content="You are an expert software developer for a software company."))
        filter_text = "all methods in this class"
        if len(method_filter) > 0:
            filter_text = "these methods in the class: %s" % str(
                method_filter)
        messages.append(chatbot.prepare_message(role="user", content="The following is a python class. I would like you to write docstrings in python for %s following the %s format. Be sure to include descriptions of all inputs (along with default values) and outputs, as well as a description of what each method does.  Class Code: %s" % (filter_text, docstring_format, class_code)))
        typer.secho("Prompting chatbot and awaiting a response ...", fg="blue")
        response = chatbot.ask(messages=messages, **
                               json.loads(generator_kwargs))
        typer.secho(
            "Successfully queried chatbot. Received response: \n\n %s" % response, fg="green")

        if output_path:
            with open(output_path, "w+") as f:
                f.write(response)
            typer.secho("Successfully wrote response out to file %s" %
                        output_path)

    else:
        typer.secho(
            "No lines found for class %s, aborting. Please check your inputs and try again." % class_name)



def load_chatbot(generator_class, **kwargs):
    typer.secho("Loading generator class %s..." % generator_class, fg="blue")
    generator_cl = utils.load_class(generator_class, "component")
    logger_cl = utils.load_class("StdOutLogger", "component")
    chatbot = generator_cl(dependent_integrations={"component": {"logger": logger_cl()}}, 
                           is_standalone=True, 
                           **kwargs)
    #suppress logging/notifying to reduce cli noise 
    chatbot.suppress_logger = True
    chatbot.suppress_notifier = True
    typer.secho("Successfully loaded generator class %s" %
                generator_class, fg="green")
    return chatbot 

def get_source_from_file(source_file, class_name): 
    typer.secho("Loading class %s from file %s and reading source code..." %
                (class_name, source_file), fg="blue")
    mod = utils.load_module_from_file(source_file)
    class_code = inspect.getsource(getattr(mod, class_name))
    num_lines = class_code.count("\n")
    typer.secho("Successfully loaded class %s! Found %s lines of code" %(class_name, num_lines), fg="green")

    return class_code, num_lines



from lolpop.component.feature_transformer.base_feature_transformer import BaseFeatureTransformer
from lolpop.utils import common_utils as utils

from pathlib import Path 


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class LocalFeatureTransformer(BaseFeatureTransformer):
    __REQUIRED_CONF__ = {"config": ["feature_transformer_path", "transformer_class"]

    }

    __DEFAULT_CONF__ = {
        "config": {"transform_func": "transform",
                   "fit_func": "fit",
                   "fit_params": {}, 
                   "transform_params": {},
                   }
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        transformer_path = Path(self._get_config("feature_transformer_path"))
        self.params = self._get_config("fit_params",{})

        if transformer_path.exists():
            transformer = utils.load_plugin(transformer_path, self)

            transformer_class = self._get_config("transformer_class")
            transformer_cl = getattr(transformer, self._get_config("transformer_class"))
            self.transformer = transformer_cl()
            self.log("Successfully loaded local transformer %s" %transformer_class)
        else:
            self.log("Transformer path not found %s" % transformer_path)

    def fit(self, X_data, y_data=None, *args, **kwargs): 
        """
        Fits the feature transformer with the given data.

        Args:
            X_data: The input data to fit the transformer.
            y_data: the label data. 
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The fitted feature transformer.

        Raises:
            Exception: If no feature transformer is found, unable to fit the transformer.
        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to fit transformer.")
        if kwargs is not None:
            self.params.update(kwargs)

        fit_func = self._get_config("fit_func")
        if hasattr(self.transformer, fit_func): 
            fit_transformer = getattr(self.transformer, fit_func)
            fit_transformer(self.config, X_data, y_data, **self.params)
        else: 
            self.notify("Unable to find fit function %s for transformer %s" %(fit_func, self._get_config("transformer_class")))

        return self.transformer

    def transform(self, data, *args, **kwargs):
        """
        Transforms the given data using the fitted feature transformer.

        Args:
            data: The input data to transform.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The transformed data.

        Raises:
            Exception: If no feature transformer is found, unable to transform features.
        """

        if not hasattr(self, "transformer"): 
            raise Exception("No feature transformer found. Unable to transform features.")
        if kwargs is not None:
            self._get_config("transform_params", {}).update(kwargs)

        transform_func = self._get_config("transform_func")
        if hasattr(self.transformer, transform_func):
            transform = getattr(self.transformer, transform_func)

            data_out = transform(self.config, data, **kwargs)
            
        else: 
            raise Exception("Unable to find transform function %s for transformer %s" % (
                transform_func, self._get_config("transformer_class")))

        return data_out 
    
    def fit_transform(self, X_data, y_data = None, *args, **kwargs): 
        """
        Fits the feature transformer with the given data and then transforms the data.

        Args:
            X_data: The input data to fit and transform.
            y_data: The label data. 
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The transformed data.

        Raises:
            Exception: If no feature transformer is found, unable to fit or transform.
        """
        self.fit(X_data, y_data, *args, **kwargs)

        data_out = self.transform(X_data, *args, **kwargs)

        return data_out  



from lolpop.component.model_explainer.base_model_explainer import BaseModelExplainer
from lolpop.utils import common_utils as utils
import alibi.explainers as alibi
import shap 
import os
from matplotlib import pyplot as plt
import numpy as np

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class AlibiModelExplainer(BaseModelExplainer): 
    __REQUIRED_CONF__ = {
        "config" : ["explainer_class", "local_dir"]
    }

    def get_explanations(self, data, model, model_version, label, classification_type=None, to_list=False, skip_explainer_plots=True, *args, **kwargs): 
        """This method generates SHAP-based feature importance scores for a given model and input data. If skip_explainer_plots is True, only the feature importance scores will be returned in Alibi format. If it is False, the method will save the SHAP summary and dependence plots as artifacts.

        Args:
        data (pandas.DataFrame): The input data to explain the model.
        model (object): The machine learning model we wish to explain.
        model_version (string): The model version object obtain from the metadata_tracker.
        label (string): The model targeta.
        classification_type (string): The type of classification. This parameter is only used if the problem type is classification. 
        to_list (boolean): If True the returned explanations will be in a list format.
        skip_explainer_plots (boolean): If True, the skip the SHAP plots.=

        Returns:
        A dictionary containing SHAP explainer and SHAP values.
        """
        #load explainer class and instatiate
        explainer_class = self._get_config("explainer_class")
        explainer_cl = getattr(alibi, explainer_class)
        explainer = explainer_cl(model._get_model(), task=self.problem_type, feature_names=data.columns.to_list())

        #fit the explainer and generate explanations
        explainer.fit()
        explanations = explainer.explain(data)

        #generate shap_plots 
        if not skip_explainer_plots:
            self._get_shap_plots(explanations.shap_values, explanations.expected_value, data, model._get_model(), label, model_version, classification_type)

        #log feature importance         
        self.metadata_tracker.log_metadata(model_version, id = "%s_global_feature_importance" %label, data = explanations.raw.get("importances").get("aggregated"))
  
        if to_list: 
            explanations = explanations.raw.get("instances").tolist() 

        return explanations

    def get_feature_importance(self, data_dict, model, model_version, *args, **kwargs): 
        """ This method generates SHAP-based feature importance scores for train and test sets. It saves SHAP summary and dependence plots as artifacts. 

        Args:
        data_dict (dict): A dictionary of the input data to explain.
        model (object): The machine learning model we wish to explain.
        model_version (string): The model version object obtain from the metadata_tracker.

        Returns:
        Tuple of two Alibi data objects:
        - the SHAP-based feature importance for the training set.
        - the SHAP-based feature importance for the test set.
        """
        #generate train/test datasets
        (train_X, train_y), (test_X, test_y) = self.data_splitter.get_train_test_dfs(data_dict, combine_xy=False) 
        if self.problem_type == "classification": 
            classification_type = utils.get_multiclass(train_y.unique())
        else: 
            classification_type = None 

        skip_explainer_plots = self._get_config("skip_explainer_plots", False)
        explanations_train = self.get_explanations(train_X, model, model_version, "train", classification_type, skip_explainer_plots=skip_explainer_plots)
        explanations_test = self.get_explanations(test_X, model, model_version, "test", classification_type, skip_explainer_plots=True)

        #compare test and train 
        expected_value_diff, feature_importance_diff = self._compare_train_test_feat_importance(explanations_train, explanations_test, classification_type)
        
        #log feature importance
        self.metadata_tracker.log_metadata(model_version, id = "feature_importance_expected_value_diff_train_test", data = expected_value_diff)
        self.metadata_tracker.log_metadata(model_version, id = "feature_importance_feature_value_diff_train_test", data = feature_importance_diff)

        return explanations_train, explanations_test

    #generate all the shap plot
    def _get_shap_plots(self, shap_values, expected_value, data, model, label, model_version, classification_type=None): 
        """This method generates various SHAP plots for the given SHAP values and saves them as artifacts.

        Args:
        shap_values (numpy.ndarray): The SHAP values to generate the plots.
        expected_value (float): The expected value of the model.
        data (pandas.DataFrame): The input data to the model.
        model (object): The machine learning model to explain.
        label (string): A label for the data.
        model_version (string): The model version.
        classification_type (string): The type of classification.

        Returns:
        None.
        """
        #note: classificaiton_type == None currently when making a prediction
        if classification_type == "multiclass":
            #bar plot w/ all classes
            shap.summary_plot(shap_values, data, show=False)
            self._save_pyplot("shap_summary_plot_bar", label, model_version)
        else:
            expected_value = [expected_value] 

        
        for i in range(len(shap_values)):
            ##Note: for multiclass, many plots only work on one class at a time.
            class_str = "" 
            if classification_type == "multiclass": 
                class_str = "_" + str(i)
            #scatter plot for class
            shap.summary_plot(shap_values[i], data, show=False)
            self._save_pyplot("shap_summary_plot_scatter%s" %class_str, label, model_version)

            #bar plot for class
            shap.summary_plot(shap_values[i], data, plot_type="bar", show=False)
            self._save_pyplot("shap_summary_plot_bar%s" %class_str, label, model_version)

            ##force plot for class
            ##force plot is very slow for large matrices, so we'll impose a sample
            sample_size = 100
            idx = np.random.randint(len(shap_values[i]), size=sample_size)
            # force plot doesn't really work w/ multiclass right now
            if classification_type !="multiclass":
                hplt = shap.force_plot(expected_value[i], shap_values[i][idx,:], data.loc[idx] ,show=False)
                force_html = "<head>%s</head> <body>%s</body>" %(shap.getjs(), hplt.html())
                self._save_file("shap_force_plot_bar%s" %class_str, label, force_html, model_version)

            #decision plots
            shap.decision_plot(expected_value[i], shap_values[i][idx,:], show=False)
            self._save_pyplot("shap_decision_plot%s" %class_str, label, model_version)

            #create dependence plots -- one for each feature
            for ft in data.columns: 
                #you can set interaction_index to determine which feature is used to color plot.
                #if omitted it will just use what seems to have the best interaction. If none it will turn off coloring
                shap.dependence_plot(ft, shap_values[i], data, interaction_index=None, show=False)
                self._save_pyplot("shap_dependence_plot%s_%s" %(class_str, ft), label, model_version)\

                #partial dependenceplots
                if i == 0: #only need to calculate these once
                    shap.partial_dependence_plot(ft, model.predict, data, show=False)
                    self._save_pyplot("shap_partial_dependence_plot_%s" %ft, label, model_version)

            # we can also check bias too if not doing it elsewhere: 
            #for bias in config.get('BIAS_COLUMNS').split(","): 
            #    shap.group_difference_plot(shap_values[i], data[bias].values==1, feature_names=data.columns)
            #    save_pyplot("shap_group_difference_plot_%s_bias_%s" %(str(i),bias), label, model_version)
            
            #waterfall_plot only shows the effect of one row at a time
            #need to construct proper exp[lanation object for this plot instead of using alibi wrapper
            shap_ex = shap._explanation.Explanation(shap_values[i], base_values=expected_value)
            shap.waterfall_plot(shap_ex[0], show=False)
            self._save_pyplot("shap_waterfall_example%s" %class_str, label, model_version)

            #heatmap
            #commenting out because this takes a long time to generate
            #shap.plots.heatmap(shap_ex)
            #self._save_pyplot("shap_heatmap%s" %class_str, label, model_version)
            
            #other plots
            #multioutput_decision_plot only shows the effect of one row at a time, but it's super awesome for multiclass problems
            
   #save shapley plut using matplotlib.pyplot
    def _save_pyplot(self, name, label, model_version): 
        """This method saves the generated SHAP plot as an artifact.

        Args:
        name (string): A name for the plot.
        label (string): The label of the data.
        model_version (string): The model version object obtained from the metadata_tracker.

        Returns:
        The saved artifact.
        """
        local_path = "%s/%s/%s" %(self._get_config("local_dir"), self.metadata_tracker.get_resource_id(model_version), label)
        os.makedirs(local_path, exist_ok=True)
        key = "%s_%s"%(name, label)
        local_file = "%s/%s.png" %(local_path, key)
        plt.savefig(local_file)
        artifact = self.metadata_tracker.log_artifact(model_version, id = key, path = local_file, external=False)
        #plt.clf()
        plt.close()

        return artifact

    def _save_file(self, name, label, content, model_version, extension="html"): 
        """This method saves the generated SHAP plot to a file as an artifact.

        Args:
        name (string): A name for the file.
        label (string): The label of the data.
        content (string): The content of the file.
        model_version (string): The version of the model to explain.
        extension (string): The extension of the file.

        Returns:
        The saved artifact.
        """
        local_path = "%s/%s/%s" % (self._get_config("local_dir"),self.metadata_tracker.get_resource_id(model_version), label)
        os.makedirs(local_path, exist_ok=True)
        key = "%s_%s" % (name, label)
        local_file = "%s/%s.%s" % (local_path, key, extension)
        with open(local_file, "w+") as f: 
            f.write(content)
        artifact = self.metadata_tracker.log_artifact(model_version, id=key, path=local_file, external=False)
 
        return artifact

    def _compare_train_test_feat_importance(self, explanations_train, explanations_test, classification_type, threshold=0.25): 
        """This method compares the SHAP-based feature importance scores between the train and test sets. 

        Args:
        explanations_train (pandas.DataFrame): The SHAP-based feature importance for the training set.
        explanations_test (pandas.DataFrame): The SHAP-based feature importance for the test set.
        classification_type (string): The type of classification.
        threshold (float): A threshold value for checking the differences between train and test sets.

        Returns:
        Tuple containing the expected differences and the feature importance differences.
        """

        #shap_train = explanations_train.shap_values
        #shap_test = explanations_test.shap_values

        #make everything 3dim so the following code is the same regardless of type
        #if classification_type !="multiclass": 
        #    shap_train = [shap_train]
        #    shap_test = [shap_test]

        #check expected values 
        if classification_type == "multiclass": 
            expected_diff = [(x - y)/max(x, 0.00001) for x,y in zip(explanations_train.expected_value, explanations_test.expected_value)]
        else: 
            expected_diff = [(explanations_train.expected_value -
                             explanations_test.expected_value)/max(explanations_train.expected_value, 0.00001)]

        for x in expected_diff: 
            if x> threshold: 
                self.notify("Threshold exceeded in difference between train/test shap expected values.")

        #check avg value for feature
        # we could recalculate everything, but shap provides the globals in the aggregated features
        #np_train = np.array(shap_train)
        #train_ft_values = np.abs(np_train).mean(axis=0).mean(axis=0)
        #np_test = np.array(shap_test)
        #test_ft_values = np.abs(np_test).mean(axis=0).mean(axis=0)

        raw_train_importances = explanations_train.raw.get("importances").get("aggregated")
        raw_test_importances = explanations_test.raw.get("importances").get("aggregated")
        # sort values to ensure features match up
        train_ft_values = sorted(
            [ x[1] for x in  zip(raw_train_importances.get("names"), raw_train_importances.get("ranked_effect"))] 
            ) 
        test_ft_values = sorted(
            [ x[1] for x in  zip(raw_test_importances.get("names"), raw_test_importances.get("ranked_effect"))]
            ) 

        feature_diff = [(x-y)/max(x, 0.00001) for x,y in zip(train_ft_values, test_ft_values)]

        for x in feature_diff: 
            if x > threshold: 
                self.notify("Threshold exceeded in difference between a feature shap value impact in train & test datasets.")

        return expected_diff, feature_diff
 



from lolpop.component.notifier.base_notifier import BaseNotifier
import base64
from email.mime.text import MIMEText
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from datetime import datetime 
import os

#note, you need to set up your google account to allow gmail api access and generate a credentials json file. 
#some instructions on that can be found here if you are lost: 
#https://developers.google.com/gmail/api/quickstart/python
#https://mailtrap.io/blog/python-send-email-gmail/#How-to-send-an-email-with-Python-via-Gmail-API
#
#Note: this works but requires browser auth due to oauth2 for the initial authorization. 
#Not sure if there's a more streamlined version to do this w/ gmail. May have to just use generic smtp server/SMTPNotificer.
class GMailNotifier(BaseNotifier): 
    __REQUIRED_CONF__ = {
        "config" : ["gcloud_credentials_file", "gcloud_credentails_token", "sender_email", "receipient_email"]
    }

    def __init__(self, *args, **kwargs): 
        #set normal config
        super().__init__(*args, **kwargs)
        creds_file = self._get_config("gcloud_credentials_file")
        token_file = self._get_config("gcloud_credentials_token")
        SCOPES = ["https://www.googleapis.com/auth/gmail.send"]

        if os.path.exists(token_file):
            creds = Credentials.from_authorized_user_file(token_file, SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    creds_file, SCOPES)
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(token_file, 'w') as token:
                token.write(creds.to_json())
        
        service = build('gmail', 'v1', credentials=creds)
        self.service = service

    def notify(self, msg, level="ERROR", *args, **kwargs):
        """
       Sends a notification email using the authenticated Gmail API service object.

       Parameters:
       ----------
       msg : str
          The message to include in the email.
       level : str, optional
          The level of the notification, default is "ERROR".
       **kwargs : dict
          A dictionary of keyword arguments, including:
             None

       Returns:
       -------
       message : object
          The sent message object.
       """
        message = MIMEText(msg)
        #message['from'] = self.email_config.get("sender_email")
        message['to'] = self._get_config("recipient_email")
        message['subject'] = "Notification from lolpop Workflow"
        create_message = {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}

        try:
            message = (self.service.users().messages().send(userId=self._get_config("sender_email"),body=create_message).execute())
            self.log('Sent notification email: %s' %message)
        except Exception as error:
            self.log('An error occurred: %s' %error, "ERROR")
            message = None


from lolpop.component.notifier.base_notifier import BaseNotifier

#mainly to be used for development/testing purposes... when you don't actually want to notify
class StdOutNotifier(BaseNotifier):
    __REQUIRED_CONF__ = {
        "config": []
    }

    def notify(self, msg, level="ERROR", *args, **kwargs):
        """
        Print the error message to standard output.

        Args:
        - msg (str): the message to be printed to standard output
        - level (str): the level of the message (default: "ERROR")
        - **kwargs: additional keyword arguments that may be ignored

        Returns:
        - None

        """
        print(msg)


from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.postgres_data_connector import PostgresDataConnector


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class PostgresDataTransformer(BaseDataTransformer):

    __REQUIRED_CONF__ = {"config": ["POSTGRES_HOST",
                                    "POSTGRES_PORT",
                                    "POSTGRES_USER",
                                    "POSTGRES_PASSWORD",
                                    "POSTGRES_DBNAME",
                                    "POSTGRES_SCHEMA"]}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pg_config = utils.load_config([
            "POSTGRES_HOST",
            "POSTGRES_PORT",
            "POSTGRES_USER",
            "POSTGRES_PASSWORD",
            "POSTGRES_DBNAME",
            "POSTGRES_SCHEMA",
        ], self.config)

    def transform(self, sql, *args, **kwargs):
        """
        Retrieves data from a PostgresDataConnector object based on SQL provided

        Args:
            sql (str): SQL query to retrieve data from the database
        Returns:
            data (pandas dataframe): A pandas dataframe containing the data retrieved from the Postgres database
        """
        data = PostgresDataConnector._load_data(
            sql, self.pg_config)
        return data



from lolpop.component.base_component import BaseComponent
from typing import Any

class BaseOrchestrator(BaseComponent):

    def decorator(self, func, cls) -> Any:
       pass

    def package(self, lolpop_class, lolpop_module, lolpop_entrypoint, *args, **kwargs) -> None:
        pass 

    def deploy(self, deployment_name, *args, **kwargs) -> None:
        pass
    
    def run(self, deployment_name, *args, **kwargs) -> tuple[str, str]:
        pass 
       
    def stop(self, deployment_name, *args, **kwargs) -> None:
        pass


from lolpop.pipeline.process.base_process import BaseProcess
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class OfflineProcess(BaseProcess): 
    __REQUIRED_CONF__ = {
        "component": ["data_transformer", "metadata_tracker", "resource_version_control", "data_profiler", "data_checker"], 
        "config": []
    }

    def transform_data(self, source_data_name, *args, **kwargs): 
        """
        Transforms source data by calling data_transformer class method.

        Args:
        source_data_name (str): A string containing the name of source data.

        Returns:
        data_out: The transformed data.
        """
        
        ##get source data
        #data = self.data_connector.get_data(source_data_name)

        #transform data
        data_out = self.data_transformer.transform(source_data_name)

        return data_out

    def track_data(self, data, id, *args, **kwargs): 
        """
        Tracks the data by creating a dataset version and registering version control metadata.

        Args:
        data (object): Data to be tracked.
        id (str): The id of the dataset version.

        Returns:
        dataset_version: The dataset version of the registered version control metadata.
        """
        #create dataset version 
        dataset_version = self.metadata_tracker.create_resource(id, type="dataset_version")
        self.datasets_used.append(dataset_version)
        
        #version data
        vc_info = self.resource_version_control.version_data(dataset_version, data)

        #register version control metadata w/ metadata tracker
        self.metadata_tracker.register_vc_resource(dataset_version, vc_info, file_type="csv")
        
        return dataset_version
      
    def profile_data(self, data, dataset_version, *args, **kwargs): 
        """
        Profiles the data by logging the data profile to metadata tracker.

        Args:
        data (object): Data to be profiled.
        dataset_version (object): The dataset version of the registered version control metadata.

        Returns:
        None
        """
        
        #profile data
        data_profile, file_path = self.data_profiler.profile_data(data)

        #log profile to metadata tracker
        self.metadata_tracker.log_data_profile(
            dataset_version, 
            file_path=file_path, 
            profile=data_profile, 
            profiler_class=type(self.data_profiler).__name__
            )

    def check_data(self, data, dataset_version, *args, **kwargs): 
        """
        Checks the data by logging a data report to the metadata tracker and sending a notification if `checks_status` is `ERROR` or `WARN`.

        Args:
        data (object): Data to be checked.
        dataset_version (object): The dataset version of the registered version control metadata.

        Returns:
        None
        """
        #run data checks
        data_report, file_path, checks_status = self.data_checker.check_data(data)

        #log data report to metadata tracker
        self.metadata_tracker.log_checks(
            dataset_version,
            file_path = file_path, 
            report = data_report, 
            checker_class = type(self.data_checker).__name__, 
            type = "data"
            )

        if checks_status == "ERROR" or checks_status == "WARN": 
            url = self.metadata_tracker.url
            self.notify("Issues found with data checks. Visit %s for more information." %url, checks_status)
        
    def compare_data(self, data, dataset_version, *args, **kwargs):
        """
        Compares a dataset version with the previous version and logs a comparison report to metadata tracker.

        Args:
        data (object): Data to be compared.
        dataset_version (object): The dataset version of the registered version control metadata.

        Returns:
        None
        """
        #get previous dataset version & dataframe 
        prev_dataset_version = self.metadata_tracker.get_prev_resource_version(dataset_version)

        if prev_dataset_version is not None: 
            vc_info = self.metadata_tracker.get_vc_info(prev_dataset_version)
            prev_data = self.resource_version_control.get_data(prev_dataset_version, vc_info)

            #compare current dataset version with previous dataset version
            if prev_data is not None and not prev_data.empty:
                comparison_report, file_path = self.data_profiler.compare_data(data, prev_data)

                self.metadata_tracker.log_data_comparison(
                    dataset_version,
                    file_path = file_path, 
                    report = comparison_report, 
                    profiler_class = type(self.data_profiler).__name__
                    )
        else: 
            self.log("No previous dataset version found for dataset: %s" %(self.metadata_tracker.get_resource_id(dataset_version)))


from lolpop.component.base_component import BaseComponent
from typing import Any

class BaseDataConnector(BaseComponent):

    def get_data(self, source, *args, **kwargs) -> Any:
        pass

    def save_data(self, data, target, *args, **kwargs):
        pass

    def _load_data(self, source, config, *args, **kwargs) -> Any: 
        pass

    def _save_data(self, data, config, *args, **kwargs): 
        pass 